"""Base visitor that implements an identity AST transform.

Subclass TransformVisitor to perform non-trivial transformations.
"""

from typing import List, Dict, cast

from mypy.nodes import (
    MypyFile, Import, Node, ImportAll, ImportFrom, FuncItem, FuncDef,
    OverloadedFuncDef, ClassDef, Decorator, Block, Var,
    OperatorAssignmentStmt, ExpressionStmt, AssignmentStmt, ReturnStmt,
    RaiseStmt, AssertStmt, DelStmt, BreakStmt, ContinueStmt,
    PassStmt, GlobalDecl, WhileStmt, ForStmt, IfStmt, TryStmt, WithStmt,
    CastExpr, TupleExpr, GeneratorExpr, ListComprehension, ListExpr,
    ConditionalExpr, DictExpr, SetExpr, NameExpr, IntExpr, StrExpr, BytesExpr,
    UnicodeExpr, FloatExpr, CallExpr, SuperExpr, MemberExpr, IndexExpr,
    SliceExpr, OpExpr, UnaryExpr, FuncExpr, TypeApplication, PrintStmt,
    SymbolTable, RefExpr, TypeVarExpr, PromoteExpr,
    ComparisonExpr, TempNode, StarExpr,
    YieldFromExpr, NamedTupleExpr, NonlocalDecl, SetComprehension,
    DictionaryComprehension, ComplexExpr, TypeAliasExpr, EllipsisExpr,
    YieldExpr, ExecStmt, Argument, BackquoteExpr
)
from mypy.types import Type, FunctionLike, Instance
from mypy.visitor import NodeVisitor


class TransformVisitor(NodeVisitor[Node]):
    """Transform a semantically analyzed AST (or subtree) to an identical copy.

    Use the node() method to transform an AST node.

    Subclass to perform a non-identity transform.

    Notes:

     * Do not duplicate TypeInfo nodes. This would generally not be desirable.
     * Only update some name binding cross-references, but only those that
       refer to Var nodes, not those targeting ClassDef, TypeInfo or FuncDef
       nodes.
     * Types are not transformed, but you can override type() to also perform
       type transformation.

    TODO nested classes and functions have not been tested well enough
    """

    def __init__(self) -> None:
        # There may be multiple references to a Var node. Keep track of
        # Var translations using a dictionary.
        self.var_map = {}  # type: Dict[Var, Var]

    def visit_mypy_file(self, node: MypyFile) -> Node:
        # NOTE: The 'names' and 'imports' instance variables will be empty!
        new = MypyFile(self.nodes(node.defs), [], node.is_bom,
                       ignored_lines=set(node.ignored_lines))
        new._name = node._name
        new._fullname = node._fullname
        new.path = node.path
        new.names = SymbolTable()
        return new

    def visit_import(self, node: Import) -> Node:
        return Import(node.ids[:])

    def visit_import_from(self, node: ImportFrom) -> Node:
        return ImportFrom(node.id, node.relative, node.names[:])

    def visit_import_all(self, node: ImportAll) -> Node:
        return ImportAll(node.id, node.relative)

    def copy_argument(self, argument: Argument) -> Argument:
        init_stmt = None  # type: AssignmentStmt

        if argument.initialization_statement:
            init_lvalue = cast(
                NameExpr,
                self.node(argument.initialization_statement.lvalues[0]),
            )
            init_lvalue.set_line(argument.line)
            init_stmt = AssignmentStmt(
                [init_lvalue],
                self.node(argument.initialization_statement.rvalue),
                self.optional_type(argument.initialization_statement.type),
            )

        arg = Argument(
            self.visit_var(argument.variable),
            argument.type_annotation,
            argument.initializer,
            argument.kind,
            init_stmt,
        )

        # Refresh lines of the inner things
        arg.set_line(argument.line)

        return arg

    def visit_func_def(self, node: FuncDef) -> FuncDef:
        # Note that a FuncDef must be transformed to a FuncDef.
        new = FuncDef(node.name(),
                      [self.copy_argument(arg) for arg in node.arguments],
                      self.block(node.body),
                      cast(FunctionLike, self.optional_type(node.type)))

        self.copy_function_attributes(new, node)

        new._fullname = node._fullname
        new.is_decorated = node.is_decorated
        new.is_conditional = node.is_conditional
        new.is_abstract = node.is_abstract
        new.is_static = node.is_static
        new.is_class = node.is_class
        new.is_property = node.is_property
        new.original_def = node.original_def
        return new

    def visit_func_expr(self, node: FuncExpr) -> Node:
        new = FuncExpr([self.copy_argument(arg) for arg in node.arguments],
                       self.block(node.body),
                       cast(FunctionLike, self.optional_type(node.type)))
        self.copy_function_attributes(new, node)
        return new

    def copy_function_attributes(self, new: FuncItem,
                                 original: FuncItem) -> None:
        new.info = original.info
        new.min_args = original.min_args
        new.max_pos = original.max_pos
        new.is_overload = original.is_overload
        new.is_generator = original.is_generator
        new.line = original.line

    def duplicate_inits(self,
                        inits: List[AssignmentStmt]) -> List[AssignmentStmt]:
        result = []  # type: List[AssignmentStmt]
        for init in inits:
            if init:
                result.append(self.duplicate_assignment(init))
            else:
                result.append(None)
        return result

    def visit_overloaded_func_def(self, node: OverloadedFuncDef) -> Node:
        items = [self.visit_decorator(decorator)
                 for decorator in node.items]
        for newitem, olditem in zip(items, node.items):
            newitem.line = olditem.line
        new = OverloadedFuncDef(items)
        new._fullname = node._fullname
        new.type = self.type(node.type)
        new.info = node.info
        return new

    def visit_class_def(self, node: ClassDef) -> Node:
        new = ClassDef(node.name,
                       self.block(node.defs),
                       node.type_vars,
                       self.nodes(node.base_type_exprs),
                       node.metaclass)
        new.fullname = node.fullname
        new.info = node.info
        new.base_types = []
        for base in node.base_types:
            new.base_types.append(cast(Instance, self.type(base)))
        new.decorators = [decorator.accept(self)
                          for decorator in node.decorators]
        new.is_builtinclass = node.is_builtinclass
        return new

    def visit_global_decl(self, node: GlobalDecl) -> Node:
        return GlobalDecl(node.names[:])

    def visit_nonlocal_decl(self, node: NonlocalDecl) -> Node:
        return NonlocalDecl(node.names[:])

    def visit_block(self, node: Block) -> Block:
        return Block(self.nodes(node.body))

    def visit_decorator(self, node: Decorator) -> Decorator:
        # Note that a Decorator must be transformed to a Decorator.
        func = self.visit_func_def(node.func)
        func.line = node.func.line
        new = Decorator(func, self.nodes(node.decorators),
                        self.visit_var(node.var))
        new.is_overload = node.is_overload
        return new

    def visit_var(self, node: Var) -> Var:
        # Note that a Var must be transformed to a Var.
        if node in self.var_map:
            return self.var_map[node]
        new = Var(node.name(), self.optional_type(node.type))
        new.line = node.line
        new._fullname = node._fullname
        new.info = node.info
        new.is_self = node.is_self
        new.is_ready = node.is_ready
        new.is_initialized_in_class = node.is_initialized_in_class
        new.is_staticmethod = node.is_staticmethod
        new.is_classmethod = node.is_classmethod
        new.is_property = node.is_property
        new.set_line(node.line)
        self.var_map[node] = new
        return new

    def visit_expression_stmt(self, node: ExpressionStmt) -> Node:
        return ExpressionStmt(self.node(node.expr))

    def visit_assignment_stmt(self, node: AssignmentStmt) -> Node:
        return self.duplicate_assignment(node)

    def duplicate_assignment(self, node: AssignmentStmt) -> AssignmentStmt:
        new = AssignmentStmt(self.nodes(node.lvalues),
                             self.node(node.rvalue),
                             self.optional_type(node.type))
        new.line = node.line
        return new

    def visit_operator_assignment_stmt(self,
                                       node: OperatorAssignmentStmt) -> Node:
        return OperatorAssignmentStmt(node.op,
                                      self.node(node.lvalue),
                                      self.node(node.rvalue))

    def visit_while_stmt(self, node: WhileStmt) -> Node:
        return WhileStmt(self.node(node.expr),
                         self.block(node.body),
                         self.optional_block(node.else_body))

    def visit_for_stmt(self, node: ForStmt) -> Node:
        return ForStmt(self.node(node.index),
                       self.node(node.expr),
                       self.block(node.body),
                       self.optional_block(node.else_body))

    def visit_return_stmt(self, node: ReturnStmt) -> Node:
        return ReturnStmt(self.optional_node(node.expr))

    def visit_assert_stmt(self, node: AssertStmt) -> Node:
        return AssertStmt(self.node(node.expr))

    def visit_del_stmt(self, node: DelStmt) -> Node:
        return DelStmt(self.node(node.expr))

    def visit_if_stmt(self, node: IfStmt) -> Node:
        return IfStmt(self.nodes(node.expr),
                      self.blocks(node.body),
                      self.optional_block(node.else_body))

    def visit_break_stmt(self, node: BreakStmt) -> Node:
        return BreakStmt()

    def visit_continue_stmt(self, node: ContinueStmt) -> Node:
        return ContinueStmt()

    def visit_pass_stmt(self, node: PassStmt) -> Node:
        return PassStmt()

    def visit_raise_stmt(self, node: RaiseStmt) -> Node:
        return RaiseStmt(self.optional_node(node.expr),
                         self.optional_node(node.from_expr))

    def visit_try_stmt(self, node: TryStmt) -> Node:
        return TryStmt(self.block(node.body),
                       self.optional_names(node.vars),
                       self.optional_nodes(node.types),
                       self.blocks(node.handlers),
                       self.optional_block(node.else_body),
                       self.optional_block(node.finally_body))

    def visit_with_stmt(self, node: WithStmt) -> Node:
        return WithStmt(self.nodes(node.expr),
                        self.optional_nodes(node.target),
                        self.block(node.body))

    def visit_print_stmt(self, node: PrintStmt) -> Node:
        return PrintStmt(self.nodes(node.args),
                         node.newline,
                         self.optional_node(node.target))

    def visit_exec_stmt(self, node: ExecStmt) -> Node:
        return ExecStmt(self.node(node.expr),
                        self.optional_node(node.variables1),
                        self.optional_node(node.variables2))

    def visit_star_expr(self, node: StarExpr) -> Node:
        return StarExpr(node.expr)

    def visit_int_expr(self, node: IntExpr) -> Node:
        return IntExpr(node.value)

    def visit_str_expr(self, node: StrExpr) -> Node:
        return StrExpr(node.value)

    def visit_bytes_expr(self, node: BytesExpr) -> Node:
        return BytesExpr(node.value)

    def visit_unicode_expr(self, node: UnicodeExpr) -> Node:
        return UnicodeExpr(node.value)

    def visit_float_expr(self, node: FloatExpr) -> Node:
        return FloatExpr(node.value)

    def visit_complex_expr(self, node: ComplexExpr) -> Node:
        return ComplexExpr(node.value)

    def visit_ellipsis(self, node: EllipsisExpr) -> Node:
        return EllipsisExpr()

    def visit_name_expr(self, node: NameExpr) -> Node:
        return self.duplicate_name(node)

    def duplicate_name(self, node: NameExpr) -> NameExpr:
        # This method is used when the transform result must be a NameExpr.
        # visit_name_expr() is used when there is no such restriction.
        new = NameExpr(node.name)
        new.info = node.info
        self.copy_ref(new, node)
        return new

    def visit_member_expr(self, node: MemberExpr) -> Node:
        member = MemberExpr(self.node(node.expr),
                            node.name)
        if node.def_var:
            member.def_var = self.visit_var(node.def_var)
        self.copy_ref(member, node)
        return member

    def copy_ref(self, new: RefExpr, original: RefExpr) -> None:
        new.kind = original.kind
        new.fullname = original.fullname
        target = original.node
        if isinstance(target, Var):
            target = self.visit_var(target)
        new.node = target
        new.is_def = original.is_def

    def visit_yield_from_expr(self, node: YieldFromExpr) -> Node:
        return YieldFromExpr(self.node(node.expr))

    def visit_yield_expr(self, node: YieldExpr) -> Node:
        return YieldExpr(self.node(node.expr))

    def visit_call_expr(self, node: CallExpr) -> Node:
        return CallExpr(self.node(node.callee),
                        self.nodes(node.args),
                        node.arg_kinds[:],
                        node.arg_names[:],
                        self.optional_node(node.analyzed))

    def visit_op_expr(self, node: OpExpr) -> Node:
        new = OpExpr(node.op, self.node(node.left), self.node(node.right))
        new.method_type = self.optional_type(node.method_type)
        return new

    def visit_comparison_expr(self, node: ComparisonExpr) -> Node:
        new = ComparisonExpr(node.operators, self.nodes(node.operands))
        new.method_types = [self.optional_type(t) for t in node.method_types]
        return new

    def visit_cast_expr(self, node: CastExpr) -> Node:
        return CastExpr(self.node(node.expr),
                        self.type(node.type))

    def visit_super_expr(self, node: SuperExpr) -> Node:
        new = SuperExpr(node.name)
        new.info = node.info
        return new

    def visit_unary_expr(self, node: UnaryExpr) -> Node:
        new = UnaryExpr(node.op, self.node(node.expr))
        new.method_type = self.optional_type(node.method_type)
        return new

    def visit_list_expr(self, node: ListExpr) -> Node:
        return ListExpr(self.nodes(node.items))

    def visit_dict_expr(self, node: DictExpr) -> Node:
        return DictExpr([(self.node(key), self.node(value))
                         for key, value in node.items])

    def visit_tuple_expr(self, node: TupleExpr) -> Node:
        return TupleExpr(self.nodes(node.items))

    def visit_set_expr(self, node: SetExpr) -> Node:
        return SetExpr(self.nodes(node.items))

    def visit_index_expr(self, node: IndexExpr) -> Node:
        new = IndexExpr(self.node(node.base), self.node(node.index))
        if node.method_type:
            new.method_type = self.type(node.method_type)
        if node.analyzed:
            if isinstance(node.analyzed, TypeApplication):
                new.analyzed = self.visit_type_application(node.analyzed)
            else:
                new.analyzed = self.visit_type_alias_expr(node.analyzed)
            new.analyzed.set_line(node.analyzed.line)
        return new

    def visit_type_application(self, node: TypeApplication) -> TypeApplication:
        return TypeApplication(self.node(node.expr),
                               self.types(node.types))

    def visit_list_comprehension(self, node: ListComprehension) -> Node:
        generator = self.duplicate_generator(node.generator)
        generator.set_line(node.generator.line)
        return ListComprehension(generator)

    def visit_set_comprehension(self, node: SetComprehension) -> Node:
        generator = self.duplicate_generator(node.generator)
        generator.set_line(node.generator.line)
        return SetComprehension(generator)

    def visit_dictionary_comprehension(self, node: DictionaryComprehension) -> Node:
        return DictionaryComprehension(self.node(node.key), self.node(node.value),
                             [self.node(index) for index in node.indices],
                             [self.node(s) for s in node.sequences],
                             [[self.node(cond) for cond in conditions]
                              for conditions in node.condlists])

    def visit_generator_expr(self, node: GeneratorExpr) -> Node:
        return self.duplicate_generator(node)

    def duplicate_generator(self, node: GeneratorExpr) -> GeneratorExpr:
        return GeneratorExpr(self.node(node.left_expr),
                             [self.node(index) for index in node.indices],
                             [self.node(s) for s in node.sequences],
                             [[self.node(cond) for cond in conditions]
                              for conditions in node.condlists])

    def visit_slice_expr(self, node: SliceExpr) -> Node:
        return SliceExpr(self.optional_node(node.begin_index),
                         self.optional_node(node.end_index),
                         self.optional_node(node.stride))

    def visit_conditional_expr(self, node: ConditionalExpr) -> Node:
        return ConditionalExpr(self.node(node.cond),
                               self.node(node.if_expr),
                               self.node(node.else_expr))

    def visit_backquote_expr(self, node: BackquoteExpr) -> Node:
        return BackquoteExpr(self.node(node.expr))

    def visit_type_var_expr(self, node: TypeVarExpr) -> Node:
        return TypeVarExpr(node.name(), node.fullname(),
                           self.types(node.values),
                           self.type(node.upper_bound), variance=node.variance)

    def visit_type_alias_expr(self, node: TypeAliasExpr) -> TypeAliasExpr:
        return TypeAliasExpr(node.type)

    def visit_namedtuple_expr(self, node: NamedTupleExpr) -> Node:
        return NamedTupleExpr(node.info)

    def visit__promote_expr(self, node: PromoteExpr) -> Node:
        return PromoteExpr(node.type)

    def visit_temp_node(self, node: TempNode) -> Node:
        return TempNode(self.type(node.type))

    def node(self, node: Node) -> Node:
        new = node.accept(self)
        new.set_line(node.line)
        return new

    # Helpers
    #
    # All the node helpers also propagate line numbers.

    def optional_node(self, node: Node) -> Node:
        if node:
            return self.node(node)
        else:
            return None

    def block(self, block: Block) -> Block:
        new = self.visit_block(block)
        new.line = block.line
        return new

    def optional_block(self, block: Block) -> Block:
        if block:
            return self.block(block)
        else:
            return None

    def nodes(self, nodes: List[Node]) -> List[Node]:
        return [self.node(node) for node in nodes]

    def optional_nodes(self, nodes: List[Node]) -> List[Node]:
        return [self.optional_node(node) for node in nodes]

    def blocks(self, blocks: List[Block]) -> List[Block]:
        return [self.block(block) for block in blocks]

    def names(self, names: List[NameExpr]) -> List[NameExpr]:
        return [self.duplicate_name(name) for name in names]

    def optional_names(self, names: List[NameExpr]) -> List[NameExpr]:
        result = []  # type: List[NameExpr]
        for name in names:
            if name:
                result.append(self.duplicate_name(name))
            else:
                result.append(None)
        return result

    def type(self, type: Type) -> Type:
        # Override this method to transform types.
        return type

    def optional_type(self, type: Type) -> Type:
        if type:
            return self.type(type)
        else:
            return None

    def types(self, types: List[Type]) -> List[Type]:
        return [self.type(type) for type in types]

    def optional_types(self, types: List[Type]) -> List[Type]:
        return [self.optional_type(type) for type in types]



# Stubs for xml.etree.ElementPath (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Tuple, List, Union, TypeVar, Callable, Optional
from .ElementTree import _ElementInterface

xpath_tokenizer_re = ... # type: Callable[..., List[Tuple[str, str]]]


class xpath_descendant_or_self: ...

_T = TypeVar('_T')

class Path:
    def __init__(self, path: str) -> None: ...
    def find(self, element: _ElementInterface) -> Optional[_ElementInterface]: ...
    def findtext(self, element: _ElementInterface, default: _T=...) -> Union[str, _T]: ...
    def findall(self, element: _ElementInterface) -> List[_ElementInterface]: ...

def find(element: _ElementInterface, path: str) -> Optional[_ElementInterface]: ...

def findtext(element: _ElementInterface, path: str, default: _T=...) -> Union[str, _T]: ...

def findall(element: _ElementInterface, path: str) -> List[_ElementInterface]: ...



# Provisional stubs for six.moves (Python 2.7)

from cStringIO import StringIO as cStringIO
from itertools import ifilter as filter
from itertools import ifilterfalse as filterfalse
from __builtin__ import raw_input as input
from __builtin__ import intern as intern
from itertools import imap as map
from os import getcwdu as getcwd
from os import getcwd as getcwdb
from __builtin__ import xrange as range
from __builtin__ import reload as reload_module
from __builtin__ import reduce as reduce
from pipes import quote as shlex_quote
from StringIO import StringIO as StringIO
from UserDict import UserDict as UserDict
from UserList import UserList as UserList
from UserString import UserString as UserString
from __builtin__ import xrange as xrange
from itertools import izip as zip
from itertools import izip_longest as zip_longest
import six.moves.cPickle as cPickle
import HTMLParser as html_parser
import htmlentitydefs as html_entities

import six.moves.urllib_parse as urllib_parse
import six.moves.urllib_error as urllib_error
import six.moves.urllib as urllib
import six.moves.urllib_robotparser as urllib_robotparser



class Annotated(object):
    def __new__(cls, *args): ...
    def __init__(self, element, values): ...
    def _annotate(self, values): ...
    def _with_annotations(self, values): ...
    def _deannotate(self, values=..., clone: bool=...): ...
    def _compiler_dispatch(self, visitor, **kw): ...
    def _constructor(self): ...
    def _clone(self): ...
    def __hash__(self): ...
    def __eq__(self): ...



# Stubs for sqlalchemy.util.compat (Python 2)

from typing import Any
from collections import namedtuple

import threading
import pickle
from six.moves.urllib.parse import (quote_plus, unquote_plus,
                                    parse_qsl, quote, unquote)
# import configparser
from six.moves import StringIO

from io import BytesIO as byte_buffer

from operator import attrgetter as dottedgetter

from six.moves import zip_longest

py33 = ... # type: Any
py32 = ... # type: Any
py3k = ... # type: Any
py2k = ... # type: Any
py265 = ... # type: Any
jython = ... # type: Any
pypy = ... # type: Any
win32 = ... # type: Any
cpython = ... # type: Any
next = ... # type: Any
safe_kwarg = ... # type: Any

ArgSpec = namedtuple('ArgSpec', ['args', 'varargs', 'keywords', 'defaults'])

def inspect_getargspec(func): ...

string_types = ... # type: Any
binary_type = ... # type: Any
text_type = unicode
int_types = ... # type: Any

def callable(fn): ...
def cmp(a, b): ...

itertools_filterfalse = ... # type: Any
itertools_filter = ... # type: Any
itertools_imap = ... # type: Any

def b64encode(x): ...
def b64decode(x): ...

def iterbytes(buf): ...
def u(s): ...
def ue(s): ...
def b(s): ...
def import_(*args): ...

reduce = ... # type: Any

def print_(*args, **kwargs): ...

time_func = ... # type: Any

def reraise(tp, value, tb=..., cause=...): ...
def raise_from_cause(exception, exc_info=...): ...

def exec_(func_text, globals_, lcl=...): ...
def with_metaclass(meta, *bases): ...
def nested(*managers): ...



from .base import Immutable, Executable, \
    ColumnCollection, ColumnSet, Generative
from .elements import ClauseElement, TextClause, ClauseList, \
    Grouping, UnaryExpression, BindParameter
from .annotation import Annotated
from .visitors import Visitable

def subquery(alias, *args, **kwargs): ...
def alias(selectable, name=..., flat: bool=...): ...

class Selectable(ClauseElement):
    def selectable(self): ...

class HasPrefixes(object):
    def prefix_with(self, *expr, **kw): ...

class HasSuffixes(object):
    def suffix_with(self, *expr, **kw): ...

class FromClause(Selectable):
    def count(self, functions, whereclause=None, **params): ...
    def select(self, whereclause=None, **params): ...
    def join(self, right, onclause=None, isouter: bool=False): ...
    def outerjoin(self, right, onclause=None): ...
    def alias(self, name=None, flat: bool=False): ...
    def is_derived_from(self, fromclause): ...
    def _is_lexical_equivalent(self, other): ...
    def replace_selectable(self, sqlutil, old, alias): ...
    def correspond_on_equivalents(self, column, equivalents): ...
    def corresponding_column(self, column, require_embedded: bool=False): ...
    @property
    def description(self): ...
    def _reset_exported(self): ...
    @property
    def columns(self): ...
    @property
    def primary_key(self): ...
    @property
    def foreign_keys(self): ...
    def _init_collections(self): ...
    @property
    def _cols_populated(self): ...
    def _populate_column_collection(self): ...
    def _refresh_for_new_column(self, column): ...

class Join(FromClause): ...
class Alias(FromClause): ...
class CTE(Generative, HasSuffixes, Alias): ...
class FromGrouping(FromClause): ...
class TableClause(Immutable, FromClause): ...

class ForUpdateArg(ClauseElement): ...
class SelectBase(Executable, FromClause): ...
class GenerativeSelect(SelectBase): ...
class CompoundSelect(GenerativeSelect): ...
class Select(HasPrefixes, HasSuffixes, GenerativeSelect): ...
class ScalarSelect(Generative, Grouping): ...
class Exists(UnaryExpression): ...
class TextAsFrom(SelectBase): ...
class AnnotatedFromClause(Annotated): ...



# TODO(MichalPokorny): better types

from typing import Any, IO, List, Optional, Union

def bindtextdomain(domain: str, localedir: str = ...) -> str: ...
def bind_textdomain_codeset(domain: str, codeset: str = ...) -> str: ...
def textdomain(domain: str = ...) -> str: ...
def gettext(message: str) -> str: ...
def lgettext(message: str) -> str: ...
def dgettext(domain: str, message: str) -> str: ...
def ldgettext(domain: str, message: str) -> str: ...
def ngettext(singular: str, plural: str, n: int) -> str: ...
def lngettext(singular: str, plural: str, n: int) -> str: ...
def dngettext(domain: str, singular: str, plural: str, n: int) -> str: ...
def ldngettext(domain: str, singular: str, plural: str, n: int) -> str: ...

class Translations(object):
    def _parse(self, fp: IO[str]) -> None: ...
    def add_fallback(self, fallback: Any) -> None: ...
    def gettext(self, message: str) -> str: ...
    def lgettext(self, message: str) -> str: ...
    def ugettext(self, message: str) -> unicode: ...
    def ngettext(self, singular: str, plural: str, n: int) -> str: ...
    def lngettext(self, singular: str, plural: str, n: int) -> str: ...
    def ungettext(self, singular: str, plural: str, n: int) -> str: ...
    def info(self) -> Any: ...
    def charset(self) -> Any: ...
    def output_charset(self) -> Any: ...
    def set_output_charset(self, charset: Any) -> None: ...
    def install(self, unicode: bool = ..., names: Any = ...) -> None: ...

# TODO: NullTranslations, GNUTranslations

def find(domain: str, localedir: str = ..., languages: List[str] = ...,
         all: Any = ...) -> Optional[Union[str, List[str]]]: ...

def translation(domain: str, localedir: str = ..., languages: List[str] = ...,
                class_: Any = ..., fallback: Any = ..., codeset: Any = ...) -> Translations: ...
def install(domain: str, localedir: str = ..., unicode: Any = ..., codeset: Any = ...,
            names: Any = ...) -> None: ...



# Stubs for ssl (Python 2)

from typing import Any
from socket import socket, error as socket_error
from collections import namedtuple

class SSLError(OSError): ...
class SSLEOFError(SSLError): ...
class SSLSyscallError(SSLError): ...
class SSLWantReadError(SSLError): ...
class SSLWantWriteError(SSLError): ...
class SSLZeroReturnError(SSLError): ...

OPENSSL_VERSION = ... # type: str
OPENSSL_VERSION_INFO = ... # type: Any
OPENSSL_VERSION_NUMBER = ... # type: int

VERIFY_CRL_CHECK_CHAIN = ... # type: int
VERIFY_CRL_CHECK_LEAF = ... # type: int
VERIFY_DEFAULT = ... # type: int
VERIFY_X509_STRICT = ... # type: int

ALERT_DESCRIPTION_ACCESS_DENIED = ... # type: int
ALERT_DESCRIPTION_BAD_CERTIFICATE = ... # type: int
ALERT_DESCRIPTION_BAD_CERTIFICATE_HASH_VALUE = ... # type: int
ALERT_DESCRIPTION_BAD_CERTIFICATE_STATUS_RESPONSE = ... # type: int
ALERT_DESCRIPTION_BAD_RECORD_MAC = ... # type: int
ALERT_DESCRIPTION_CERTIFICATE_EXPIRED = ... # type: int
ALERT_DESCRIPTION_CERTIFICATE_REVOKED = ... # type: int
ALERT_DESCRIPTION_CERTIFICATE_UNKNOWN = ... # type: int
ALERT_DESCRIPTION_CERTIFICATE_UNOBTAINABLE = ... # type: int
ALERT_DESCRIPTION_CLOSE_NOTIFY = ... # type: int
ALERT_DESCRIPTION_DECODE_ERROR = ... # type: int
ALERT_DESCRIPTION_DECOMPRESSION_FAILURE = ... # type: int
ALERT_DESCRIPTION_DECRYPT_ERROR = ... # type: int
ALERT_DESCRIPTION_HANDSHAKE_FAILURE = ... # type: int
ALERT_DESCRIPTION_ILLEGAL_PARAMETER = ... # type: int
ALERT_DESCRIPTION_INSUFFICIENT_SECURITY = ... # type: int
ALERT_DESCRIPTION_INTERNAL_ERROR = ... # type: int
ALERT_DESCRIPTION_NO_RENEGOTIATION = ... # type: int
ALERT_DESCRIPTION_PROTOCOL_VERSION = ... # type: int
ALERT_DESCRIPTION_RECORD_OVERFLOW = ... # type: int
ALERT_DESCRIPTION_UNEXPECTED_MESSAGE = ... # type: int
ALERT_DESCRIPTION_UNKNOWN_CA = ... # type: int
ALERT_DESCRIPTION_UNKNOWN_PSK_IDENTITY = ... # type: int
ALERT_DESCRIPTION_UNRECOGNIZED_NAME = ... # type: int
ALERT_DESCRIPTION_UNSUPPORTED_CERTIFICATE = ... # type: int
ALERT_DESCRIPTION_UNSUPPORTED_EXTENSION = ... # type: int
ALERT_DESCRIPTION_USER_CANCELLED = ... # type: int

OP_ALL = ... # type: int
OP_CIPHER_SERVER_PREFERENCE = ... # type: int
OP_NO_COMPRESSION = ... # type: int
OP_NO_SSLv2 = ... # type: int
OP_NO_SSLv3 = ... # type: int
OP_NO_TLSv1 = ... # type: int
OP_NO_TLSv1_1 = ... # type: int
OP_NO_TLSv1_2 = ... # type: int
OP_SINGLE_DH_USE = ... # type: int
OP_SINGLE_ECDH_USE = ... # type: int

SSL_ERROR_EOF = ... # type: int
SSL_ERROR_INVALID_ERROR_CODE = ... # type: int
SSL_ERROR_SSL = ... # type: int
SSL_ERROR_SYSCALL = ... # type: int
SSL_ERROR_WANT_CONNECT = ... # type: int
SSL_ERROR_WANT_READ = ... # type: int
SSL_ERROR_WANT_WRITE = ... # type: int
SSL_ERROR_WANT_X509_LOOKUP = ... # type: int
SSL_ERROR_ZERO_RETURN = ... # type: int

CERT_NONE = ... # type: int
CERT_OPTIONAL = ... # type: int
CERT_REQUIRED = ... # type: int

PROTOCOL_SSLv23 = ... # type: int
PROTOCOL_SSLv3 = ... # type: int
PROTOCOL_TLSv1 = ... # type: int
PROTOCOL_TLSv1_1 = ... # type: int
PROTOCOL_TLSv1_2 = ... # type: int

HAS_ECDH = ... # type: bool
HAS_NPN = ... # type: bool
HAS_SNI = ... # type: bool

CHANNEL_BINDING_TYPES = ... # type: Any

class CertificateError(ValueError): ...

def match_hostname(cert, hostname): ...

DefaultVerifyPaths = namedtuple('DefaultVerifyPaths', 'cafile capath openssl_cafile_env openssl_cafile openssl_capath_env openssl_capath')

def get_default_verify_paths(): ...

class _ASN1Object:
    def __new__(cls, oid): ...
    @classmethod
    def fromnid(cls, nid): ...
    @classmethod
    def fromname(cls, name): ...

class Purpose(_ASN1Object):
    SERVER_AUTH = ... # type: Purpose
    CLIENT_AUTH = ... # type: Purpose

class _SSLContext:
    check_hostname = ... # type: Any
    options = ... # type: Any
    verify_flags = ... # type: Any
    verify_mode = ... # type: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def _set_npn_protocols(self, *args, **kwargs): ...
    def _wrap_socket(self, *args, **kwargs): ...
    def cert_store_stats(self): ...
    def get_ca_certs(self, binary_form=...): ...
    def load_cert_chain(self, *args, **kwargs): ...
    def load_dh_params(self, *args, **kwargs): ...
    def load_verify_locations(self, *args, **kwargs): ...
    def session_stats(self, *args, **kwargs): ...
    def set_ciphers(self, *args, **kwargs): ...
    def set_default_verify_paths(self, *args, **kwargs): ...
    def set_ecdh_curve(self, *args, **kwargs): ...
    def set_servername_callback(self, method): ...

class SSLContext(_SSLContext):
    def __new__(cls, protocol, *args, **kwargs): ...
    protocol = ... # type: Any
    def __init__(self, protocol): ...
    def wrap_socket(self, sock, server_side=False, do_handshake_on_connect=True, suppress_ragged_eofs=True, server_hostname=None): ...
    def set_npn_protocols(self, npn_protocols): ...
    def set_alpn_protocols(self, alpn_protocols): ...
    def load_default_certs(self, purpose=...): ...

def create_default_context(purpose=..., cafile=None, capath=None, cadata=None): ...

class SSLSocket(socket):
    keyfile = ... # type: Any
    certfile = ... # type: Any
    cert_reqs = ... # type: Any
    ssl_version = ... # type: Any
    ca_certs = ... # type: Any
    ciphers = ... # type: Any
    server_side = ... # type: Any
    server_hostname = ... # type: Any
    do_handshake_on_connect = ... # type: Any
    suppress_ragged_eofs = ... # type: Any
    def __init__(self, sock=None, keyfile=None, certfile=None, server_side=False, cert_reqs=..., ssl_version=..., ca_certs=None, do_handshake_on_connect=True, family=..., type=..., proto=0, fileno=None, suppress_ragged_eofs=True, npn_protocols=None, ciphers=None, server_hostname=None, _context=None): ...
    @property
    def context(self): ...
    @context.setter
    def context(self, ctx): ...
    def dup(self): ...
    def read(self, len=0, buffer=None): ...
    def write(self, data): ...
    def getpeercert(self, binary_form=False): ...
    def selected_npn_protocol(self): ...
    def selected_alpn_protocol(self): ...
    def cipher(self): ...
    def compression(self): ...
    def send(self, data, flags=0): ...
    def sendto(self, data, flags_or_addr, addr=None): ...
    def sendall(self, data, flags=0): ...
    def recv(self, buflen=1024, flags=0): ...
    def recv_into(self, buffer, nbytes=None, flags=0): ...
    def recvfrom(self, buflen=1024, flags=0): ...
    def recvfrom_into(self, buffer, nbytes=None, flags=0): ...
    def pending(self): ...
    def shutdown(self, how): ...
    def close(self): ...
    def unwrap(self): ...
    def do_handshake(self, block=False): ...
    def connect(self, addr): ...
    def connect_ex(self, addr): ...
    def accept(self): ...
    def makefile(self, mode='', bufsize=-1): ...
    def get_channel_binding(self, cb_type=''): ...
    def version(self): ...

def wrap_socket(sock, keyfile=None, certfile=None, server_side=False, cert_reqs=..., ssl_version=..., ca_certs=None, do_handshake_on_connect=True, suppress_ragged_eofs=True, ciphers=None): ...
def cert_time_to_seconds(cert_time): ...

PEM_HEADER = ... # type: Any
PEM_FOOTER = ... # type: Any

def DER_cert_to_PEM_cert(der_cert_bytes): ...
def PEM_cert_to_DER_cert(pem_cert_string): ...
def get_server_certificate(addr, ssl_version=..., ca_certs=None): ...
def get_protocol_name(protocol_code): ...
def sslwrap_simple(sock, keyfile=None, certfile=None): ...



# Stubs for email.mime.base (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# import message

# TODO
#class MIMEBase(message.Message):
class MIMEBase:
    def __init__(self, _maintype, _subtype, **_params) -> None: ...



# Stubs for tempfile
# Ron Murawski <ron@horizonchess.com>

# based on http://docs.python.org/3.3/library/tempfile.html
# Adapted for Python 2.7 by Michal Pokorny

from typing import Tuple, IO

# global variables
tempdir = ...  # type: str
template = ...  # type: str

# TODO text files

# function stubs
def TemporaryFile(
            mode: str = ..., bufsize: int = ..., suffix: str = ...,
            prefix: str = ..., dir: str = ...) -> IO[str]: ...
def NamedTemporaryFile(
            mode: str = ..., bufsize: int = ..., suffix: str = ...,
            prefix: str = ..., dir: str = ..., delete: bool = ...
            ) -> IO[str]: ...
def SpooledTemporaryFile(
           max_size: int = ..., mode: str = ..., buffering: int = ...,
           suffix: str = ..., prefix: str = ..., dir: str = ...) -> IO[str]:
    ...

class TemporaryDirectory:
    name = ...  # type: basestring
    def __init__(self, suffix: basestring = ..., prefix: basestring = ...,
                 dir: basestring = ...) -> None: ...
    def cleanup(self) -> None: ...
    def __enter__(self) -> basestring: ...
    def __exit__(self, type, value, traceback) -> bool: ...

def mkstemp(suffix: basestring = ..., prefix: basestring = ..., dir: basestring = ...,
            text: bool = ...) -> Tuple[int, basestring]: ...
def mkdtemp(suffix: basestring = ..., prefix: basestring = ...,
            dir: basestring = ...) -> basestring: ...
def mktemp(suffix: basestring = ..., prefix: basestring = ..., dir: basestring = ...) -> basestring: ...
def gettempdir() -> basestring: ...
def gettempprefix() -> basestring: ...



# Stubs for collections

# Based on http://docs.python.org/2.7/library/collections.html

# TODO more abstract base classes (interfaces in mypy)

# NOTE: These are incomplete!

from typing import (
    Dict, Generic, TypeVar, Iterable, Tuple, Callable, Mapping, overload, Iterator, Sized,
    Optional, List, Set, Sequence, Union, Reversible, MutableMapping, MutableSequence
)
import typing

_T = TypeVar('_T')
_KT = TypeVar('_KT')
_VT = TypeVar('_VT')

# namedtuple is special-cased in the type checker; the initializer is ignored.
namedtuple = object()

class deque(Sized, Iterable[_T], Reversible[_T], Generic[_T]):
    def __init__(self, iterable: Iterable[_T] = ...,
                 maxlen: int = ...) -> None: ...
    @property
    def maxlen(self) -> Optional[int]: ...
    def append(self, x: _T) -> None: ...
    def appendleft(self, x: _T) -> None: ...
    def clear(self) -> None: ...
    def count(self, x: _T) -> int: ...
    def extend(self, iterable: Iterable[_T]) -> None: ...
    def extendleft(self, iterable: Iterable[_T]) -> None: ...
    def pop(self) -> _T: ...
    def popleft(self) -> _T: ...
    def remove(self, value: _T) -> None: ...
    def reverse(self) -> None: ...
    def rotate(self, n: int) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_T]: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...
    def __getitem__(self, i: int) -> _T: ...
    def __setitem__(self, i: int, x: _T) -> None: ...
    def __contains__(self, o: _T) -> bool: ...
    def __reversed__(self) -> Iterator[_T]: ...

class Counter(Dict[_T, int], Generic[_T]):
    # TODO: __init__ keyword arguments
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, Mapping: Mapping[_T, int]) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[_T]) -> None: ...
    def elements(self) -> Iterator[_T]: ...
    def most_common(self, n: int = ...) -> List[_T]: ...
    @overload
    def subtract(self, mapping: Mapping[_T, int]) -> None: ...
    @overload
    def subtract(self, iterable: Iterable[_T]) -> None: ...
    # The Iterable[Tuple[...]] argument type is not actually desirable
    # (the tuples will be added as keys, breaking type safety) but
    # it's included so that the signature is compatible with
    # Dict.update. Not sure if we should use '# type: ignore' instead
    # and omit the type from the union.
    @overload
    def update(self, m: Mapping[_T, int]) -> None: ...
    @overload
    def update(self, m: Union[Iterable[_T], Iterable[Tuple[_T, int]]]) -> None: ...

class OrderedDict(Dict[_KT, _VT], Generic[_KT, _VT]):
    def popitem(self, last: bool = ...) -> Tuple[_KT, _VT]: ...
    def move_to_end(self, key: _KT, last: bool = ...) -> None: ...

class defaultdict(Dict[_KT, _VT], Generic[_KT, _VT]):
    default_factory = ...  # type: Callable[[], _VT]
    # TODO: __init__ keyword args
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, map: Mapping[_KT, _VT]) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[Tuple[_KT, _VT]]) -> None: ...
    @overload
    def __init__(self, default_factory: Callable[[], _VT]) -> None: ...
    @overload
    def __init__(self, default_factory: Callable[[], _VT],
                 map: Mapping[_KT, _VT]) -> None: ...
    @overload
    def __init__(self, default_factory: Callable[[], _VT],
                 iterable: Iterable[Tuple[_KT, _VT]]) -> None: ...
    def __missing__(self, key: _KT) -> _VT: ...



"""Type operations"""

from mypy.types import Type, AnyType, TypeTranslator, TypeVarType


def replace_type_vars(typ: Type, func_tvars: bool = True) -> Type:
    """Replace type variable references in a type with the Any type. If
    func_tvars is false, only replace instance type variables.
    """
    return typ.accept(ReplaceTypeVarsVisitor(func_tvars))


class ReplaceTypeVarsVisitor(TypeTranslator):
    # Only override type variable handling; otherwise perform an indentity
    # transformation.

    func_tvars = False

    def __init__(self, func_tvars: bool) -> None:
        self.func_tvars = func_tvars

    def visit_type_var(self, t: TypeVarType) -> Type:
        if t.id > 0 or self.func_tvars:
            if t.line is not None:
                return AnyType(t.line)
            else:
                return AnyType()
        else:
            return t


def replace_func_type_vars(typ: Type, target_type: Type) -> Type:
    """Replace function type variables in a type with the target type."""
    return typ.accept(ReplaceFuncTypeVarsVisitor(target_type))


class ReplaceFuncTypeVarsVisitor(TypeTranslator):
    def __init__(self, target_type: Type) -> None:
        self.target_type = target_type

    def visit_type_var(self, t: TypeVarType) -> Type:
        if t.id < 0:
            return self.target_type
        else:
            return t



# Stubs for email.feedparser (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

class BufferedSubFile:
    def __init__(self) -> None: ...
    def push_eof_matcher(self, pred): ...
    def pop_eof_matcher(self): ...
    def close(self): ...
    def readline(self): ...
    def unreadline(self, line): ...
    def push(self, data): ...
    def pushlines(self, lines): ...
    def __iter__(self): ...
    def __next__(self): ...

class FeedParser:
    policy = ...  # type: Any
    def __init__(self, _factory=..., *, policy=...) -> None: ...
    def feed(self, data): ...
    def close(self): ...

class BytesFeedParser(FeedParser):
    def feed(self, data): ...



from typing import Any, IO, AnyStr, Callable, Tuple, List
from types import TracebackType, FrameType

ExtractTbResult = List[Tuple[str, int, str, str]]

def print_tb(traceback: TracebackType, limit: int = ..., file: IO[str] = ...) -> None: ...
def print_exception(type: type, value: Exception, limit: int = ..., file: IO[str] = ...) -> None: ...
def print_exc(limit: int = ..., file: IO[str] = ...) -> None: ...
def format_exc(limit: int = ...) -> str: ...
def print_last(limit: int = ..., file: IO[str] = ...) -> None: ...
def print_stack(f: FrameType = ..., limit: int = ..., file: IO[AnyStr] = ...) -> None: ...
def extract_tb(f: TracebackType, limit: int = ...) -> ExtractTbResult: ...
def extract_stack(f: FrameType = ..., limit: int = ...) -> ExtractTbResult: ...
def format_list(list: ExtractTbResult) -> List[str]: ...
def format_exception_only(type: type, value: List[str]) -> str: ...
def format_exception(type: type, value: List[str], tb: TracebackType, limit: int = ...) -> str: ...
def format_tb(f: TracebackType, limit: int = ...) -> List[str]: ...
def format_stack(f: FrameType = ..., limit: int = ...) -> List[str]: ...
def tb_lineno(tb: TracebackType) -> AnyStr: ...




from .base import Executable, ColumnCollection
from .elements import ClauseList, Cast, Extract, _literal_as_binds, \
    literal_column, _type_from_args, ColumnElement, _clone,\
    Over, BindParameter, FunctionFilter
from .selectable import FromClause, Select, Alias
from .visitors import VisitableType

class FunctionElement(Executable, ColumnElement, FromClause): ...

class _FunctionGenerator(object):
    def __init__(self, **opts): ...
    def __getattr__(self, name): ...
    def __call__(self, *c, **kwargs) -> Function: ...

func = ... # type: _FunctionGenerator
modifier = ... # type: _FunctionGenerator

class Function(FunctionElement): ...

class _GenericMeta(VisitableType): ...
# TODO: Use GenericFunction(util.with_metaclass(_GenericMeta, Function))
class GenericFunction(_GenericMeta, Function): ...
class next_value(GenericFunction): ...

class AnsiFunction(GenericFunction): ...
class ReturnTypeFromArgs(GenericFunction): ...

class coalesce(ReturnTypeFromArgs): ...
class max(ReturnTypeFromArgs): ...
class min(ReturnTypeFromArgs): ...
class sum(ReturnTypeFromArgs): ...
class now(GenericFunction): ...
class concat(GenericFunction): ...

class char_length(GenericFunction): ...
class random(GenericFunction): ...
class count(GenericFunction): ...
class current_date(AnsiFunction): ...
class current_time(AnsiFunction): ...
class current_timestamp(AnsiFunction): ...
class current_user(AnsiFunction): ...
class localtime(AnsiFunction): ...
class localtimestamp(AnsiFunction): ...
class session_user(AnsiFunction): ...
class sysdate(AnsiFunction): ...
class user(AnsiFunction): ...



# Stubs for xml.etree.ElementTree (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, AnyStr, Union, IO, Callable, Dict, List, Tuple, Sequence, Iterator, TypeVar, Optional, KeysView, ItemsView, Generator
import io

VERSION = ... # type: str

_Ss = TypeVar('_Ss', str, bytes)
_T = TypeVar('_T')
_str_or_bytes = Union[str, bytes]

class _ElementInterface:
    tag = ... # type: _str_or_bytes
    attrib = ... # type: Dict[_str_or_bytes, _str_or_bytes]
    text = ... # type: Optional[_str_or_bytes]
    tail = ... # type: Optional[_str_or_bytes]
    def __init__(self, tag: Union[AnyStr, Callable[..., '_ElementInterface']], attrib: Dict[AnyStr, AnyStr]) -> None: ...
    def makeelement(self, tag: _Ss, attrib: Dict[_Ss, _Ss]) -> '_ElementInterface': ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> '_ElementInterface': ...
    def __setitem__(self, index: int, element: '_ElementInterface') -> None: ...
    def __delitem__(self, index: int) -> None: ...
    def __getslice__(self, start: int, stop: int) -> Sequence['_ElementInterface']: ...
    def __setslice__(self, start: int, stop: int, elements: Sequence['_ElementInterface']) -> None: ...
    def __delslice__(self, start: int, stop: int) -> None: ...
    def append(self, element: '_ElementInterface') -> None: ...
    def insert(self, index: int, element: '_ElementInterface') -> None: ...
    def remove(self, element: '_ElementInterface') -> None: ...
    def getchildren(self) -> List['_ElementInterface']: ...
    def find(self, path: str) -> Optional['_ElementInterface']: ...
    def findtext(self, path: str, default: _T=...) -> Union[str, _T]: ...
    def findall(self, path: str) -> List['_ElementInterface']: ...
    def clear(self) -> None: ...
    def get(self, key: AnyStr, default: _T=...) -> Union[AnyStr, _T]: ...
    def set(self, key: AnyStr, value: AnyStr) -> None: ...
    def keys(self) -> KeysView[AnyStr]: ...
    def items(self) -> ItemsView[AnyStr, AnyStr]: ...
    def getiterator(self, tag: Union[str, AnyStr]=...) -> List['_ElementInterface']: ...

def Element(tag: Union[AnyStr, Callable[..., _ElementInterface]], attrib: Dict[AnyStr, AnyStr]=..., **extra: Dict[str, AnyStr]) -> _ElementInterface: ...
def SubElement(parent: _ElementInterface, tag: AnyStr, attrib: Dict[AnyStr, AnyStr]=..., **extra: Dict[str, AnyStr]) -> _ElementInterface: ...
def Comment(text: _str_or_bytes=...) -> _ElementInterface: ...
def ProcessingInstruction(target: str, text: str=...) -> _ElementInterface: ...

PI = ... # type: Callable[..., _ElementInterface]

class QName:
    text = ... # type: str
    def __init__(self, text_or_uri: str, tag: str=...) -> None: ...


_file_or_filename = Union[str, bytes, int, IO[Any]]

class ElementTree:
    def __init__(self, element: _ElementInterface=..., file: _file_or_filename=...) -> None: ...
    def getroot(self) -> _ElementInterface: ...
    def parse(self, source: _file_or_filename, parser: 'XMLTreeBuilder'=...) -> _ElementInterface: ...
    def getiterator(self, tag: Union[str, AnyStr]=...) -> List[_ElementInterface]: ...
    def find(self, path: str) -> Optional[_ElementInterface]: ...
    def findtext(self, path: str, default: _T=...) -> Union[_T, str]: ...
    def findall(self, path: str) -> List[_ElementInterface]: ...
    def write(self, file_or_filename: _file_or_filename, encoding: str=...) -> None: ...

def iselement(element: _ElementInterface) -> bool: ...
def dump(elem: _ElementInterface) -> None: ...
def fixtag(tag: Union[str, QName], namespaces: Dict[str, str]) -> Tuple[str, Optional[str]]: ...
def parse(source: _file_or_filename, parser: 'XMLTreeBuilder'=...) -> ElementTree: ...


class iterparse:
    def __init__(self, source: _file_or_filename, events: Sequence[str]=...) -> None: ...
    # TODO-figure out this type...
    def __next__(self) -> Tuple[str, _ElementInterface]: ...

def XML(text: AnyStr) -> _ElementInterface: ...
def XMLID(text: AnyStr) -> Tuple[_ElementInterface, Dict[str, _ElementInterface]]: ...

# TODO-improve this type
fromstring = ... # type: Callable[..., _ElementInterface]

def tostring(element: _ElementInterface, encoding: str=...) -> AnyStr: ...

class TreeBuilder:
    def __init__(self, element_factory: Callable[[AnyStr, Dict[AnyStr, AnyStr]], _ElementInterface]=...) -> None: ...
    def close(self) -> _ElementInterface: ...
    def data(self, data: AnyStr) -> None: ...
    def start(self, tag: AnyStr, attrs: Dict[AnyStr, AnyStr]) -> _ElementInterface: ...
    def end(self, tag: AnyStr) -> _ElementInterface: ...

class XMLTreeBuilder:
    # TODO-what is entity used for???
    entity = ... # type: Any
    def __init__(self, html: int=..., target: TreeBuilder=...) -> None: ...
    def doctype(self, name: str, pubid: str, system: str) -> None: ...
    def close(self) -> Any: ...  # TODO-most of the time, this will be Element, but it can be anything target.close() returns
    def feed(self, data: AnyStr)-> None: ...



#!/usr/bin/env python

import glob
import os
import os.path
import sys

from distutils.core import setup
from mypy.version import __version__
from mypy import git

if sys.version_info < (3, 2, 0):
    sys.stderr.write("ERROR: You need Python 3.2 or later to use mypy.\n")
    exit(1)

git.verify_git_integrity_or_abort(".")

version = __version__
description = 'Optional static typing for Python'
long_description = '''
Mypy -- Optional Static Typing for Python
=========================================

Add type annotations to your Python programs, and use mypy to type
check them.  Mypy is essentially a Python linter on steroids, and it
can catch many programming errors by analyzing your program, without
actually having to run it.  Mypy has a powerful type system with
features such as type inference, gradual typing, generics and union
types.
'''.lstrip()


def find_data_files(base, globs):
    """Find all interesting data files, for setup(data_files=)

    Arguments:
      root:  The directory to search in.
      globs: A list of glob patterns to accept files.
    """

    rv_dirs = [root for root, dirs, files in os.walk(base)]
    rv = []
    for rv_dir in rv_dirs:
        files = []
        for pat in globs:
            files += glob.glob(os.path.join(rv_dir, pat))
        if not files:
            continue
        target = os.path.join('lib', 'mypy', rv_dir)
        rv.append((target, files))

    return rv

data_files = []

data_files += find_data_files('typeshed', ['*.py', '*.pyi'])

data_files += find_data_files('xml', ['*.xsd', '*.xslt', '*.css'])

classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Topic :: Software Development',
]

package_dir = {'mypy': 'mypy'}
if sys.version_info < (3, 5, 0):
    package_dir[''] = 'lib-typing/3.2'

scripts = ['scripts/mypy', 'scripts/stubgen']
if os.name == 'nt':
    scripts.append('scripts/mypy.bat')

setup(name='mypy-lang',
      version=version,
      description=description,
      long_description=long_description,
      author='Jukka Lehtosalo',
      author_email='jukka.lehtosalo@iki.fi',
      url='http://www.mypy-lang.org/',
      license='MIT License',
      platforms=['POSIX'],
      package_dir=package_dir,
      py_modules=['typing'] if sys.version_info < (3, 5, 0) else [],
      packages=['mypy'],
      scripts=scripts,
      data_files=data_files,
      classifiers=classifiers,
      )



# Stubs for io

# Based on http://docs.python.org/3.2/library/io.html

from typing import List, BinaryIO, TextIO, IO, overload, Iterator, Iterable, Any
import builtins
import codecs
import _io

DEFAULT_BUFFER_SIZE = 0  # type: int
SEEK_SET = ...  # type: int
SEEK_CUR = ...  # type: int
SEEK_END = ...  # type: int

open = builtins.open

class BlockingIOError(OSError): ...
class UnsupportedOperation(ValueError, OSError): ...

class IncrementalNewlineDecoder(codecs.IncrementalDecoder):
    newlines = ...  # type: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def decode(self, input, final=...): ...
    def getstate(self): ...
    def reset(self): ...
    def setstate(self, state): ...

class IOBase(_io._IOBase): ...
class RawIOBase(_io._RawIOBase, IOBase): ...
class BufferedIOBase(_io._BufferedIOBase, IOBase): ...
class TextIOBase(_io._TextIOBase, IOBase): ...

class FileIO(_io._RawIOBase):
    closefd = ...  # type: Any
    mode = ...  # type: Any
    def __init__(self, name, mode=..., closefd=..., opener=...) -> None: ...
    def readinto(self, b): ...
    def write(self, b): ...

class BufferedReader(_io._BufferedIOBase):
    mode = ...  # type: Any
    name = ...  # type: Any
    raw = ...  # type: Any
    def __init__(self, raw, buffer_size=...) -> None: ...
    def peek(self, size: int = ...): ...

class BufferedWriter(_io._BufferedIOBase):
    mode = ...  # type: Any
    name = ...  # type: Any
    raw = ...  # type: Any
    def __init__(self, raw, buffer_size=...) -> None: ...

class BufferedRWPair(_io._BufferedIOBase):
    def __init__(self, reader, writer, buffer_size=...) -> None: ...
    def peek(self, size: int = ...): ...

class BufferedRandom(_io._BufferedIOBase):
    mode = ...  # type: Any
    name = ...  # type: Any
    raw = ...  # type: Any
    def __init__(self, raw, buffer_size=...) -> None: ...
    def peek(self, size: int = ...): ...

class BytesIO(BinaryIO):
    def __init__(self, initial_bytes: bytes = ...) -> None: ...
    # TODO getbuffer
    # TODO see comments in BinaryIO for missing functionality
    def close(self) -> None: ...
    @property
    def closed(self) -> bool: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def read(self, n: int = ...) -> bytes: ...
    def readable(self) -> bool: ...
    def readline(self, limit: int = ...) -> bytes: ...
    def readlines(self, hint: int = ...) -> List[bytes]: ...
    def seek(self, offset: int, whence: int = ...) -> int: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def truncate(self, size: int = ...) -> int: ...
    def writable(self) -> bool: ...
    @overload
    def write(self, s: bytes) -> int: ...
    @overload
    def write(self, s: bytearray) -> int: ...
    def writelines(self, lines: Iterable[bytes]) -> None: ...
    def getvalue(self) -> bytes: ...
    def read1(self) -> str: ...

    def __iter__(self) -> Iterator[bytes]: ...
    def __enter__(self) -> 'BytesIO': ...
    def __exit__(self, t: type = None, value: BaseException = None, traceback: Any = None) -> bool: ...

class StringIO(TextIO):
    def __init__(self, initial_value: str = ...,
                 newline: str = ...) -> None: ...
    # TODO see comments in BinaryIO for missing functionality
    def close(self) -> None: ...
    @property
    def closed(self) -> bool: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def read(self, n: int = ...) -> str: ...
    def readable(self) -> bool: ...
    def readline(self, limit: int = ...) -> str: ...
    def readlines(self, hint: int = ...) -> List[str]: ...
    def seek(self, offset: int, whence: int = ...) -> int: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def truncate(self, size: int = ...) -> int: ...
    def writable(self) -> bool: ...
    def write(self, s: str) -> int: ...
    def writelines(self, lines: Iterable[str]) -> None: ...
    def getvalue(self) -> str: ...

    def __iter__(self) -> Iterator[str]: ...
    def __enter__(self) -> 'StringIO': ...
    def __exit__(self, t: type = None, value: BaseException = None, traceback: Any = None) -> bool: ...

class TextIOWrapper(TextIO):
    # TODO: This is actually a base class of _io._TextIOBase.
    # write_through is undocumented but used by subprocess
    def __init__(self, buffer: IO[bytes], encoding: str = ...,
                 errors: str = ..., newline: str = ...,
                 line_buffering: bool = ...,
                 write_through: bool = ...) -> None: ...
    # TODO see comments in BinaryIO for missing functionality
    def close(self) -> None: ...
    @property
    def closed(self) -> bool: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def read(self, n: int = ...) -> str: ...
    def readable(self) -> bool: ...
    def readline(self, limit: int = ...) -> str: ...
    def readlines(self, hint: int = ...) -> List[str]: ...
    def seek(self, offset: int, whence: int = ...) -> int: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def truncate(self, size: int = ...) -> int: ...
    def writable(self) -> bool: ...
    def write(self, s: str) -> int: ...
    def writelines(self, lines: Iterable[str]) -> None: ...

    def __iter__(self) -> Iterator[str]: ...
    def __enter__(self) -> StringIO: ...
    def __exit__(self, t: type = None, value: BaseException = None, traceback: Any = None) -> bool: ...



# Stubs for multiprocessing.pool

# NOTE: These are incomplete!

from typing import Any, Callable, Iterable, List, Sequence

class AsyncResult():
    def get(self, timeout: float = -1) -> Any: ...
    def wait(self, timeout: float = -1) -> None: ...
    def ready(self) -> bool: ...
    def successful(self) -> bool: ...

class ThreadPool():
    def __init__(self, processes: int = ...) -> None: ...
    def apply_async(self, func: Callable[..., Any],
                    args: Sequence[Any]=(),
                    kwds: Dict[str, Any]={},
                    callback: Callable[..., None] = None) -> AsyncResult: ...
    def apply(self, func: Callable[..., Any],
              args: Sequence[Any]=(),
              kwds: Dict[str, Any]={}) -> Any: ...
    def map(self, func: Callable[..., Any],
            iterable: Iterable[Any]=()) -> List[Any]: ...
    def map_async(self, func: Callable[..., Any],
                  iterable: Iterable[Any] = (),
                  chunksize: int = -1,
                  callback: Callable[..., None] = None) -> AsyncResult: ...
    def imap(self, func: Callable[..., Any],
             iterable: Iterable[Any]=()) -> Iterable[Any]: ...
    def imap_async(self, func: Callable[..., Any],
                   chunksize: int = -1,
                   iterable: Iterable[Any]=(),
                   callback: Callable[..., None] = None) -> AsyncResult: ...
    def close(self) -> None: ...
    def terminate(self) -> None: ...
    def join(self) -> None: ...



# Stubs for tokenize (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Union, TextIO
from builtins import open as _builtin_open
from token import *

COMMENT = ... # type: Any
NL = ... # type: Any
ENCODING = ... # type: Any

class TokenInfo:
    @property
    def exact_type(self): ...

class TokenError(Exception): ...
class StopTokenizing(Exception): ...

class Untokenizer:
    tokens = ... # type: Any
    prev_row = ... # type: Any
    prev_col = ... # type: Any
    encoding = ... # type: Any
    def __init__(self): ...
    def add_whitespace(self, start): ...
    def untokenize(self, iterable): ...
    def compat(self, token, iterable): ...

def untokenize(iterable): ...
def detect_encoding(readline): ...
def tokenize(readline): ...

def open(filename: Union[str, bytes, int]) -> TextIO: ...

# Names in __all__ with no definition:
#   AMPER
#   AMPEREQUAL
#   ASYNC
#   AT
#   ATEQUAL
#   AWAIT
#   CIRCUMFLEX
#   CIRCUMFLEXEQUAL
#   COLON
#   COMMA
#   DEDENT
#   DOT
#   DOUBLESLASH
#   DOUBLESLASHEQUAL
#   DOUBLESTAR
#   DOUBLESTAREQUAL
#   ELLIPSIS
#   ENDMARKER
#   EQEQUAL
#   EQUAL
#   ERRORTOKEN
#   GREATER
#   GREATEREQUAL
#   INDENT
#   ISEOF
#   ISNONTERMINAL
#   ISTERMINAL
#   LBRACE
#   LEFTSHIFT
#   LEFTSHIFTEQUAL
#   LESS
#   LESSEQUAL
#   LPAR
#   LSQB
#   MINEQUAL
#   MINUS
#   NAME
#   NEWLINE
#   NOTEQUAL
#   NT_OFFSET
#   NUMBER
#   N_TOKENS
#   OP
#   PERCENT
#   PERCENTEQUAL
#   PLUS
#   PLUSEQUAL
#   RARROW
#   RBRACE
#   RIGHTSHIFT
#   RIGHTSHIFTEQUAL
#   RPAR
#   RSQB
#   SEMI
#   SLASH
#   SLASHEQUAL
#   STAR
#   STAREQUAL
#   STRING
#   TILDE
#   VBAR
#   VBAREQUAL
#   tok_name



# Stubs for select

# NOTE: These are incomplete!

from typing import Any, Tuple, List, Sequence

class error(Exception): ...

POLLIN = 0
POLLPRI = 0
POLLOUT = 0
POLLERR = 0
POLLHUP = 0
POLLNVAL = 0

class poll:
    def __init__(self) -> None: ...
    def register(self, fd: Any,
                 eventmask: int = ...) -> None: ...
    def modify(self, fd: Any, eventmask: int) -> None: ...
    def unregister(self, fd: Any) -> None: ...
    def poll(self, timeout: int = ...) -> List[Tuple[int, int]]: ...

def select(rlist: Sequence, wlist: Sequence, xlist: Sequence,
           timeout: float = ...) -> Tuple[List[int],
                                           List[int],
                                           List[int]]: ...



# Stubs for array

# Based on http://docs.python.org/3.2/library/array.html

from typing import Any, Iterable, Tuple, List, Iterator, BinaryIO, overload

typecodes = ...  # type: str

class array:
    typecode = ...  # type: str
    itemsize = ...  # type: int
    def __init__(self, typecode: str,
                 initializer: Iterable[Any] = ...) -> None: ...
    def append(self, x: Any) -> None: ...
    def buffer_info(self) -> Tuple[int, int]: ...
    def byteswap(self) -> None: ...
    def count(self, x: Any) -> int: ...
    def extend(self, iterable: Iterable[Any]) -> None: ...
    def frombytes(self, s: bytes) -> None: ...
    def fromfile(self, f: BinaryIO, n: int) -> None: ...
    def fromlist(self, list: List[Any]) -> None: ...
    def fromstring(self, s: bytes) -> None: ...
    def fromunicode(self, s: str) -> None: ...
    def index(self, x: Any) -> int: ...
    def insert(self, i: int, x: Any) -> None: ...
    def pop(self, i: int = ...) -> Any: ...
    def remove(self, x: Any) -> None: ...
    def reverse(self) -> None: ...
    def tobytes(self) -> bytes: ...
    def tofile(self, f: BinaryIO) -> None: ...
    def tolist(self) -> List[Any]: ...
    def tostring(self) -> bytes: ...
    def tounicode(self) -> str: ...

    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...

    @overload
    def __getitem__(self, i: int) -> Any: ...
    @overload
    def __getitem__(self, s: slice) -> 'array': ...

    def __setitem__(self, i: int, o: Any) -> None: ...
    def __delitem__(self, i: int) -> None: ...
    def __add__(self, x: 'array') -> 'array': ...
    def __mul__(self, n: int) -> 'array': ...
    def __contains__(self, o: object) -> bool: ...



# Stub for pdb (incomplete, only some global functions)

from typing import Any, Dict

def run(statement: str,
        globals: Dict[str, Any] = None,
        locals: Dict[str, Any] = None) -> None:
    ...

def runeval(expression: str,
            globals: Dict[str, Any] = None,
            locals: Dict[str, Any] = None) -> Any:
    ...

def runctx(statement: str,
        globals: Dict[str, Any],
        locals: Dict[str, Any]) -> None:
    ...

def runcall(*args: Any, **kwds: Any) -> Any:
    ...

def set_trace() -> None:
    ...

def post_mortem(t: Any = None) -> None:
    ...

def pm() -> None:
    ...



# Stubs for ipaddress (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

IPV4LENGTH = ... # type: Any
IPV6LENGTH = ... # type: Any

class AddressValueError(ValueError): ...
class NetmaskValueError(ValueError): ...

def ip_address(address): ...
def ip_network(address, strict=...): ...
def ip_interface(address): ...
def v4_int_to_packed(address): ...
def v6_int_to_packed(address): ...
def summarize_address_range(first, last): ...
def collapse_addresses(addresses): ...
def get_mixed_type_key(obj): ...

class _TotalOrderingMixin:
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __gt__(self, other): ...
    def __ge__(self, other): ...

class _IPAddressBase(_TotalOrderingMixin):
    @property
    def exploded(self): ...
    @property
    def compressed(self): ...
    @property
    def version(self): ...

class _BaseAddress(_IPAddressBase):
    def __init__(self, address) -> None: ...
    def __int__(self): ...
    def __eq__(self, other): ...
    def __lt__(self, other): ...
    def __add__(self, other): ...
    def __sub__(self, other): ...
    def __hash__(self): ...

class _BaseNetwork(_IPAddressBase):
    def __init__(self, address) -> None: ...
    def hosts(self): ...
    def __iter__(self): ...
    def __getitem__(self, n): ...
    def __lt__(self, other): ...
    def __eq__(self, other): ...
    def __hash__(self): ...
    def __contains__(self, other): ...
    def overlaps(self, other): ...
    @property
    def broadcast_address(self): ...
    @property
    def hostmask(self): ...
    @property
    def with_prefixlen(self): ...
    @property
    def with_netmask(self): ...
    @property
    def with_hostmask(self): ...
    @property
    def num_addresses(self): ...
    @property
    def prefixlen(self): ...
    def address_exclude(self, other): ...
    def compare_networks(self, other): ...
    def subnets(self, prefixlen_diff=..., new_prefix=...): ...
    def supernet(self, prefixlen_diff=..., new_prefix=...): ...
    @property
    def is_multicast(self): ...
    @property
    def is_reserved(self): ...
    @property
    def is_link_local(self): ...
    @property
    def is_private(self): ...
    @property
    def is_global(self): ...
    @property
    def is_unspecified(self): ...
    @property
    def is_loopback(self): ...

class _BaseV4:
    def __init__(self, address) -> None: ...
    @property
    def max_prefixlen(self): ...
    @property
    def version(self): ...

class IPv4Address(_BaseV4, _BaseAddress):
    def __init__(self, address) -> None: ...
    @property
    def packed(self): ...
    @property
    def is_reserved(self): ...
    @property
    def is_private(self): ...
    @property
    def is_multicast(self): ...
    @property
    def is_unspecified(self): ...
    @property
    def is_loopback(self): ...
    @property
    def is_link_local(self): ...

class IPv4Interface(IPv4Address):
    network = ... # type: Any
    netmask = ... # type: Any
    hostmask = ... # type: Any
    def __init__(self, address) -> None: ...
    def __eq__(self, other): ...
    def __lt__(self, other): ...
    def __hash__(self): ...
    @property
    def ip(self): ...
    @property
    def with_prefixlen(self): ...
    @property
    def with_netmask(self): ...
    @property
    def with_hostmask(self): ...

class IPv4Network(_BaseV4, _BaseNetwork):
    network_address = ... # type: Any
    netmask = ... # type: Any
    hosts = ... # type: Any
    def __init__(self, address, strict=...) -> None: ...
    @property
    def is_global(self): ...

class _BaseV6:
    def __init__(self, address) -> None: ...
    @property
    def max_prefixlen(self): ...
    @property
    def version(self): ...

class IPv6Address(_BaseV6, _BaseAddress):
    def __init__(self, address) -> None: ...
    @property
    def packed(self): ...
    @property
    def is_multicast(self): ...
    @property
    def is_reserved(self): ...
    @property
    def is_link_local(self): ...
    @property
    def is_site_local(self): ...
    @property
    def is_private(self): ...
    @property
    def is_global(self): ...
    @property
    def is_unspecified(self): ...
    @property
    def is_loopback(self): ...
    @property
    def ipv4_mapped(self): ...
    @property
    def teredo(self): ...
    @property
    def sixtofour(self): ...

class IPv6Interface(IPv6Address):
    network = ... # type: Any
    netmask = ... # type: Any
    hostmask = ... # type: Any
    def __init__(self, address) -> None: ...
    def __eq__(self, other): ...
    def __lt__(self, other): ...
    def __hash__(self): ...
    @property
    def ip(self): ...
    @property
    def with_prefixlen(self): ...
    @property
    def with_netmask(self): ...
    @property
    def with_hostmask(self): ...
    @property
    def is_unspecified(self): ...
    @property
    def is_loopback(self): ...

class IPv6Network(_BaseV6, _BaseNetwork):
    network_address = ... # type: Any
    netmask = ... # type: Any
    def __init__(self, address, strict=...) -> None: ...
    def hosts(self): ...
    @property
    def is_site_local(self): ...



# Stubs for multiprocessing (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from multiprocessing.process import Process as Process, current_process as current_process, active_children as active_children
from multiprocessing.util import SUBDEBUG as SUBDEBUG, SUBWARNING as SUBWARNING

class ProcessError(Exception): ...
class BufferTooShort(ProcessError): ...
class TimeoutError(ProcessError): ...
class AuthenticationError(ProcessError): ...

def Manager(): ...
def Pipe(duplex=True): ...
def cpu_count(): ...
def freeze_support(): ...
def get_logger(): ...
def log_to_stderr(level=None): ...
def allow_connection_pickling(): ...
def Lock(): ...
def RLock(): ...
def Condition(lock=None): ...
def Semaphore(value=1): ...
def BoundedSemaphore(value=1): ...
def Event(): ...
def Queue(maxsize=0): ...
def JoinableQueue(maxsize=0): ...
def Pool(processes=None, initializer=None, initargs=..., maxtasksperchild=None): ...
def RawValue(typecode_or_type, *args): ...
def RawArray(typecode_or_type, size_or_initializer): ...
def Value(typecode_or_type, *args, **kwds): ...
def Array(typecode_or_type, size_or_initializer, **kwds): ...



# Python 2.7 ast

import typing
from typing import Any, Tuple, Iterator, Union

from _ast import (
    Add, alias, And, arguments, Assert, Assign, AST, Attribute, AugAssign,
    AugLoad, AugStore, BinOp, BitAnd, BitOr, BitXor, BoolOp, boolop, Break,
    Call, ClassDef, cmpop, Compare, comprehension, Continue, Del, Delete, Dict,
    DictComp, Div, Ellipsis, Eq, ExceptHandler, Exec, Expr, expr, Expression,
    expr_context, ExtSlice, FloorDiv, For, FunctionDef, GeneratorExp, Global,
    Gt, GtE, If, IfExp, Import, ImportFrom, In, Index, Interactive, Invert, Is,
    IsNot, keyword, Lambda, List, ListComp, Load, LShift, Lt, LtE, Mod, mod,
    Module, Mult, Name, Not, NotEq, NotIn, Num, operator, Or, Param, Pass, Pow,
    Print, Raise, Repr, Return, RShift, Set, SetComp, Slice, slice, stmt,
    Store, Str, Sub, Subscript, Suite, TryExcept, TryFinally, Tuple, UAdd,
    UnaryOp, unaryop, USub, While, With, Yield
)

__version__ = ...  # type: str
PyCF_ONLY_AST = ...  # type: int


def parse(source: Union[str, bytes], filename: Union[str, bytes] = ..., mode: str = ...) -> AST: ...
def copy_location(new_node: AST, old_node: AST) -> AST: ...
def dump(node: AST, annotate_fields: bool = ..., include_attributes: bool = ...) -> str: ...
def fix_missing_locations(node: AST) -> AST: ...
def get_docstring(node: AST, clean: bool = ...) -> str: ...
def increment_lineno(node: AST, n: int = ...) -> AST: ...
def iter_child_nodes(node: AST) -> Iterator[AST]: ...
def iter_fields(node: AST) -> Iterator[typing.Tuple[str, Any]]: ...
def literal_eval(node_or_string: Union[str, AST]) -> Any: ...
def walk(node: AST) -> Iterator[AST]: ...

class NodeVisitor():
    __doc__ = ...  # type: str
    def visit(self, node: AST) -> Any: ...
    def generic_visit(self, node: AST) -> None: ...

class NodeTransformer(NodeVisitor):
    __doc__ = ...  # type: str
    def generic_visit(self, node: AST) -> None: ...






# Stubs for thrift.transport.TTransport (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any
from thrift.Thrift import TException

class TTransportException(TException):
    UNKNOWN = ... # type: Any
    NOT_OPEN = ... # type: Any
    ALREADY_OPEN = ... # type: Any
    TIMED_OUT = ... # type: Any
    END_OF_FILE = ... # type: Any
    type = ... # type: Any
    def __init__(self, type=..., message=...) -> None: ...

class TTransportBase:
    def isOpen(self): ...
    def open(self): ...
    def close(self): ...
    def read(self, sz): ...
    def readAll(self, sz): ...
    def write(self, buf): ...
    def flush(self): ...

class CReadableTransport:
    @property
    def cstringio_buf(self): ...
    def cstringio_refill(self, partialread, reqlen): ...

class TServerTransportBase:
    def listen(self): ...
    def accept(self): ...
    def close(self): ...

class TTransportFactoryBase:
    def getTransport(self, trans): ...

class TBufferedTransportFactory:
    def getTransport(self, trans): ...

class TBufferedTransport(TTransportBase, CReadableTransport):
    DEFAULT_BUFFER = ... # type: Any
    def __init__(self, trans, rbuf_size=...) -> None: ...
    def isOpen(self): ...
    def open(self): ...
    def close(self): ...
    def read(self, sz): ...
    def write(self, buf): ...
    def flush(self): ...
    @property
    def cstringio_buf(self): ...
    def cstringio_refill(self, partialread, reqlen): ...

class TMemoryBuffer(TTransportBase, CReadableTransport):
    def __init__(self, value=...) -> None: ...
    def isOpen(self): ...
    def open(self): ...
    def close(self): ...
    def read(self, sz): ...
    def write(self, buf): ...
    def flush(self): ...
    def getvalue(self): ...
    @property
    def cstringio_buf(self): ...
    def cstringio_refill(self, partialread, reqlen): ...

class TFramedTransportFactory:
    def getTransport(self, trans): ...

class TFramedTransport(TTransportBase, CReadableTransport):
    def __init__(self, trans) -> None: ...
    def isOpen(self): ...
    def open(self): ...
    def close(self): ...
    def read(self, sz): ...
    def readFrame(self): ...
    def write(self, buf): ...
    def flush(self): ...
    @property
    def cstringio_buf(self): ...
    def cstringio_refill(self, prefix, reqlen): ...

class TFileObjectTransport(TTransportBase):
    fileobj = ... # type: Any
    def __init__(self, fileobj) -> None: ...
    def isOpen(self): ...
    def close(self): ...
    def read(self, sz): ...
    def write(self, buf): ...
    def flush(self): ...

class TSaslClientTransport(TTransportBase, CReadableTransport):
    START = ... # type: Any
    OK = ... # type: Any
    BAD = ... # type: Any
    ERROR = ... # type: Any
    COMPLETE = ... # type: Any
    transport = ... # type: Any
    sasl = ... # type: Any
    def __init__(self, transport, host, service, mechanism=..., **sasl_kwargs) -> None: ...
    def open(self): ...
    def send_sasl_msg(self, status, body): ...
    def recv_sasl_msg(self): ...
    def write(self, data): ...
    def flush(self): ...
    def read(self, sz): ...
    def close(self): ...
    @property
    def cstringio_buf(self): ...
    def cstringio_refill(self, prefix, reqlen): ...



# Stubs for mimetypes (Python 3.5)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

class MimeTypes:
    encodings_map = ... # type: Any
    suffix_map = ... # type: Any
    types_map = ... # type: Any
    types_map_inv = ... # type: Any
    def __init__(self, filenames=..., strict=True): ...
    def add_type(self, type, ext, strict=True): ...
    def guess_type(self, url, strict=True): ...
    def guess_all_extensions(self, type, strict=True): ...
    def guess_extension(self, type, strict=True): ...
    def read(self, filename, strict=True): ...
    def readfp(self, fp, strict=True): ...
    def read_windows_registry(self, strict=True): ...

def guess_type(url, strict=True): ...
def guess_all_extensions(type, strict=True): ...
def guess_extension(type, strict=True): ...
def add_type(type, ext, strict=True): ...
def init(files=None): ...
def read_mime_types(file): ...



# Stubs for email (Python 3.4)

from typing import IO, Any

def message_from_string(s: str, *args, **kwargs): ...
def message_from_bytes(s: bytes, *args, **kwargs): ...
def message_from_file(fp: IO[str], *args, **kwargs): ...
def message_from_binary_file(fp: IO[bytes], *args, **kwargs): ...

# Names in __all__ with no definition:
#   base64mime
#   charset
#   encoders
#   errors
#   feedparser
#   generator
#   header
#   iterators
#   message
#   mime
#   parser
#   quoprimime
#   utils



from typing import Dict, Generic, Mapping, TypeVar

_KT = TypeVar('_KT')
_VT = TypeVar('_VT')

class UserDict(Dict[_KT, _VT], Generic[_KT, _VT]):
    data = ... # type: Mapping[_KT, _VT]

    def __init__(self, initialdata: Mapping[_KT, _VT] = ...) -> None: ...

    # TODO: DictMixin



# Stubs for subprocess

# Based on http://docs.python.org/2/library/subprocess.html and Python 3 stub

from typing import Sequence, Any, Mapping, Callable, Tuple, IO, Union, Optional

_FILE = Union[int, IO[Any]]

# TODO force keyword arguments
# TODO more keyword arguments (from Popen)
def call(args: Sequence[str], *,
         stdin: _FILE = ..., stdout: _FILE = ..., stderr: _FILE = ...,
         shell: bool = ..., env: Mapping[str, str] = ...,
         cwd: str = ...) -> int: ...
def check_call(args: Sequence[str], *,
               stdin: _FILE = ..., stdout: _FILE = ..., stderr: _FILE = ...,
               shell: bool = ..., env: Mapping[str, str] = ..., cwd: str = ...,
               close_fds: Sequence[_FILE] = ..., preexec_fn: Callable[[], Any] = ...) -> int: ...
def check_output(args: Sequence[str], *,
                 stdin: _FILE = ..., stderr: _FILE = ...,
                 shell: bool = ..., universal_newlines: bool = ...,
                 env: Mapping[str, str] = ..., cwd: str = ...) -> str: ...

PIPE = ... # type: int
STDOUT = ... # type: int

class CalledProcessError(Exception):
    returncode = 0
    cmd = ...  # type: str
    output = ...  # type: str # May be None

    def __init__(self, returncode: int, cmd: str, output: str = ...) -> None: ...

class Popen:
    stdin = ... # type: Optional[IO[Any]]
    stdout = ... # type: Optional[IO[Any]]
    stderr = ... # type: Optional[IO[Any]]
    pid = 0
    returncode = 0

    def __init__(self,
                 args: Sequence[str],
                 bufsize: int = ...,
                 executable: str = ...,
                 stdin: _FILE = ...,
                 stdout: _FILE = ...,
                 stderr: _FILE = ...,
                 preexec_fn: Callable[[], Any] = ...,
                 close_fds: bool = ...,
                 shell: bool = ...,
                 cwd: str = ...,
                 env: Mapping[str, str] = ...,
                 universal_newlines: bool = ...,
                 startupinfo: Any = ...,
                 creationflags: int = ...) -> None: ...

    def poll(self) -> int: ...
    def wait(self) -> int: ...
    # Return str/bytes
    def communicate(self, input: Union[str, unicode] = ...) -> Tuple[str, str]: ...
    def send_signal(self, signal: int) -> None: ...
    def terminate(self) -> None: ...
    def kill(self) -> None: ...
    def __enter__(self) -> 'Popen': ...
    def __exit__(self, type, value, traceback) -> bool: ...

def getstatusoutput(cmd: str) -> Tuple[int, str]: ...
def getoutput(cmd: str) -> str: ...

# Windows-only: STARTUPINFO etc.

STD_INPUT_HANDLE = ... # type: Any
STD_OUTPUT_HANDLE = ... # type: Any
STD_ERROR_HANDLE = ... # type: Any
SW_HIDE = ... # type: Any
STARTF_USESTDHANDLES = ... # type: Any
STARTF_USESHOWWINDOW = ... # type: Any
CREATE_NEW_CONSOLE = ... # type: Any
CREATE_NEW_PROCESS_GROUP = ... # type: Any



from typing import Any, Callable, Generic, Optional, TypeVar

_T = TypeVar('_T')

class CallableProxyType(object):  # "weakcallableproxy"
  pass

class ProxyType(object):  # "weakproxy"
  pass

class ReferenceType(Generic[_T]):
    # TODO rest of members
    def __init__(self, o: _T, callback: Callable[[ReferenceType[_T]],
                                                 Any] = ...) -> None: ...
    def __call__(self) -> Optional[_T]: ...

ref = ReferenceType

def getweakrefcount(object: Any) -> int: ...
def getweakrefs(object: Any) -> int: ...
def proxy(object: Any, callback: Callable[[Any], Any] = ...) -> None: ...



# Stubs for wsgiref.validate (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

class WSGIWarning(Warning): ...

def validator(application): ...

class InputWrapper:
    input = ... # type: Any
    def __init__(self, wsgi_input): ...
    def read(self, *args): ...
    def readline(self): ...
    def readlines(self, *args): ...
    def __iter__(self): ...
    def close(self): ...

class ErrorWrapper:
    errors = ... # type: Any
    def __init__(self, wsgi_errors): ...
    def write(self, s): ...
    def flush(self): ...
    def writelines(self, seq): ...
    def close(self): ...

class WriteWrapper:
    writer = ... # type: Any
    def __init__(self, wsgi_writer): ...
    def __call__(self, s): ...

class PartialIteratorWrapper:
    iterator = ... # type: Any
    def __init__(self, wsgi_iterator): ...
    def __iter__(self): ...

class IteratorWrapper:
    original_iterator = ... # type: Any
    iterator = ... # type: Any
    closed = ... # type: Any
    check_start_response = ... # type: Any
    def __init__(self, wsgi_iterator, check_start_response): ...
    def __iter__(self): ...
    def next(self): ...
    def close(self): ...
    def __del__(self): ...



# Stubs for _posixsubprocess

# NOTE: These are incomplete!

from typing import Tuple, Sequence

def cloexec_pipe() -> Tuple[int, int]: ...
def fork_exec(args: Sequence[str],
              executable_list, close_fds, fds_to_keep, cwd: str, env_list,
              p2cread: int, p2cwrite: int, c2pred: int, c2pwrite: int,
              errread: int, errwrite: int, errpipe_read: int,
              errpipe_write: int, restore_signals, start_new_session,
              preexec_fn) -> int: ...



# Stubs for email.mime.message (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from email.mime.nonmultipart import MIMENonMultipart

class MIMEMessage(MIMENonMultipart):
    def __init__(self, _msg, _subtype=...) -> None: ...



# Stubs for binascii

# Based on http://docs.python.org/3.2/library/binascii.html

import typing

def a2b_uu(string: bytes) -> bytes: ...
def b2a_uu(data: bytes) -> bytes: ...
def a2b_base64(string: bytes) -> bytes: ...
def b2a_base64(data: bytes) -> bytes: ...
def a2b_qp(string: bytes, header: bool = ...) -> bytes: ...
def b2a_qp(data: bytes, quotetabs: bool = ..., istext: bool = ...,
             header: bool = ...) -> bytes: ...
def a2b_hqx(string: bytes) -> bytes: ...
def rledecode_hqx(data: bytes) -> bytes: ...
def rlecode_hqx(data: bytes) -> bytes: ...
def b2a_hqx(data: bytes) -> bytes: ...
def crc_hqx(data: bytes, crc: int) -> int: ...
def crc32(data: bytes, crc: int = ...) -> int: ...
def b2a_hex(data: bytes) -> bytes: ...
def hexlify(data: bytes) -> bytes: ...
def a2b_hex(hexstr: bytes) -> bytes: ...
def unhexlify(hexlify: bytes) -> bytes: ...

class Error(Exception): ...
class Incomplete(Exception): ...



# Stubs for email.mime.base (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from email import message

class MIMEBase(message.Message):
    def __init__(self, _maintype, _subtype, **_params) -> None: ...



# Stubs for email.parser (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any
import email.feedparser

FeedParser = email.feedparser.FeedParser
BytesFeedParser = email.feedparser.BytesFeedParser

class Parser:
    policy = ...  # type: Any
    def __init__(self, _class=..., *, policy=...) -> None: ...
    def parse(self, fp, headersonly=...): ...
    def parsestr(self, text, headersonly=...): ...

class HeaderParser(Parser):
    def parse(self, fp, headersonly=...): ...
    def parsestr(self, text, headersonly=...): ...

class BytesParser:
    parser = ...  # type: Any
    def __init__(self, *args, **kw) -> None: ...
    def parse(self, fp, headersonly=...): ...
    def parsebytes(self, text, headersonly=...): ...

class BytesHeaderParser(BytesParser):
    def parse(self, fp, headersonly=...): ...
    def parsebytes(self, text, headersonly=...): ...



# Stubs for requests.packages (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

class VendorAlias:
    def __init__(self, package_names) -> None: ...
    def find_module(self, fullname, path=...): ...
    def load_module(self, name): ...



"""Stubs for the 'select' module."""

from typing import Any, Optional, Tuple, Iterable, List

EPOLLERR = ...  # type: int
EPOLLET = ...  # type: int
EPOLLHUP = ...  # type: int
EPOLLIN = ...  # type: int
EPOLLMSG = ...  # type: int
EPOLLONESHOT = ...  # type: int
EPOLLOUT = ...  # type: int
EPOLLPRI = ...  # type: int
EPOLLRDBAND = ...  # type: int
EPOLLRDNORM = ...  # type: int
EPOLLWRBAND = ...  # type: int
EPOLLWRNORM = ...  # type: int
EPOLL_RDHUP = ...  # type: int
KQ_EV_ADD = ...  # type: int
KQ_EV_CLEAR = ...  # type: int
KQ_EV_DELETE = ...  # type: int
KQ_EV_DISABLE = ...  # type: int
KQ_EV_ENABLE = ...  # type: int
KQ_EV_EOF = ...  # type: int
KQ_EV_ERROR = ...  # type: int
KQ_EV_FLAG1 = ...  # type: int
KQ_EV_ONESHOT = ...  # type: int
KQ_EV_SYSFLAGS = ...  # type: int
KQ_FILTER_AIO = ...  # type: int
KQ_FILTER_NETDEV = ...  # type: int
KQ_FILTER_PROC = ...  # type: int
KQ_FILTER_READ = ...  # type: int
KQ_FILTER_SIGNAL = ...  # type: int
KQ_FILTER_TIMER = ...  # type: int
KQ_FILTER_VNODE = ...  # type: int
KQ_FILTER_WRITE = ...  # type: int
KQ_NOTE_ATTRIB = ...  # type: int
KQ_NOTE_CHILD = ...  # type: int
KQ_NOTE_DELETE = ...  # type: int
KQ_NOTE_EXEC = ...  # type: int
KQ_NOTE_EXIT = ...  # type: int
KQ_NOTE_EXTEND = ...  # type: int
KQ_NOTE_FORK = ...  # type: int
KQ_NOTE_LINK = ...  # type: int
KQ_NOTE_LINKDOWN = ...  # type: int
KQ_NOTE_LINKINV = ...  # type: int
KQ_NOTE_LINKUP = ...  # type: int
KQ_NOTE_LOWAT = ...  # type: int
KQ_NOTE_PCTRLMASK = ...  # type: int
KQ_NOTE_PDATAMASK = ...  # type: int
KQ_NOTE_RENAME = ...  # type: int
KQ_NOTE_REVOKE = ...  # type: int
KQ_NOTE_TRACK = ...  # type: int
KQ_NOTE_TRACKERR = ...  # type: int
KQ_NOTE_WRITE = ...  # type: int
PIPE_BUF = ...  # type: int
POLLERR = ...  # type: int
POLLHUP = ...  # type: int
POLLIN = ...  # type: int
POLLMSG = ...  # type: int
POLLNVAL = ...  # type: int
POLLOUT = ...  # type: int
POLLPRI = ...  # type: int
POLLRDBAND = ...  # type: int
POLLRDNORM = ...  # type: int
POLLWRBAND = ...  # type: int
POLLWRNORM = ...  # type: int

def poll() -> epoll: ...
def select(rlist, wlist, xlist, timeout: float = None) -> Tuple[List, List, List]: ...

class error(Exception): ...

class kevent(object):
    data = ... # type: Any
    fflags = ... # type: int
    filter = ... # type: int
    flags = ... # type: int
    ident = ... # type: Any
    udata = ... # type: Any
    def __init__(self, *args, **kwargs) -> None: ...

class kqueue(object):
    closed = ... # type: bool
    def __init__(self) -> None: ...
    def close(self) -> None: ...
    def control(self, changelist: Optional[Iterable[kevent]], max_events: int, timeout: int = ...) -> List[kevent]: ...
    def fileno(self) -> int: ...
    @classmethod
    def fromfd(cls, fd: int) -> kqueue: ...

class epoll(object):
    def __init__(self, sizehint: int = ...) -> None: ...
    def close(self) -> None: ...
    def fileno(self) -> int: ...
    def register(self, fd: int, eventmask: int = ...) -> None: ...
    def modify(self, fd: int, eventmask: int) -> None: ...
    def unregister(fd: int) -> None: ...
    def poll(timeout: float = ..., maxevents: int = ...) -> Any: ...
    @classmethod
    def fromfd(self, fd: int) -> epoll: ...



"""Stub file for the '_operator' module."""
# This is an autogenerated file. It serves as a starting point
# for a more precise manual annotation of this module.
# Feel free to edit the source below, but remove this header when you do.

from typing import Any, List, Tuple, Dict, Generic

def _compare_digest(a, b) -> bool:
    raise BufferError()
    raise TypeError()

def abs(*args, **kwargs) -> Any: ...

def add(*args, **kwargs) -> Any: ...

def and_(*args, **kwargs) -> Any: ...

def concat(*args, **kwargs) -> Any: ...

def contains(*args, **kwargs) -> bool: ...

def countOf(*args, **kwargs) -> int: ...

def delitem(*args, **kwargs) -> None: ...

def eq(*args, **kwargs) -> Any: ...

def floordiv(*args, **kwargs) -> Any: ...

def ge(*args, **kwargs) -> Any: ...

def getitem(*args, **kwargs) -> Any: ...

def gt(*args, **kwargs) -> Any: ...

def iadd(*args, **kwargs) -> Any: ...

def iand(*args, **kwargs) -> Any: ...

def iconcat(*args, **kwargs) -> Any: ...

def ifloordiv(*args, **kwargs) -> Any: ...

def ilshift(*args, **kwargs) -> Any: ...

def imod(*args, **kwargs) -> Any: ...

def imul(*args, **kwargs) -> Any: ...

def index(*args, **kwargs) -> Any: ...

def indexOf(*args, **kwargs) -> int: ...

def inv(*args, **kwargs) -> Any: ...

def invert(*args, **kwargs) -> Any: ...

def ior(*args, **kwargs) -> Any: ...

def ipow(*args, **kwargs) -> Any: ...

def irshift(*args, **kwargs) -> Any: ...

def is_(*args, **kwargs) -> bool: ...

def is_not(*args, **kwargs) -> bool: ...

def isub(*args, **kwargs) -> Any: ...

def itruediv(*args, **kwargs) -> Any: ...

def ixor(*args, **kwargs) -> Any: ...

def le(*args, **kwargs) -> Any: ...

def length_hint(a, *args, **kwargs) -> int: ...

def lshift(*args, **kwargs) -> Any: ...

def lt(*args, **kwargs) -> Any: ...

def mod(*args, **kwargs) -> Any: ...

def mul(*args, **kwargs) -> Any: ...

def ne(*args, **kwargs) -> Any: ...

def neg(*args, **kwargs) -> Any: ...

def not_(*args, **kwargs) -> bool: ...

def or_(*args, **kwargs) -> Any: ...

def pos(*args, **kwargs) -> Any: ...

def pow(*args, **kwargs) -> Any: ...

def rshift(*args, **kwargs) -> Any: ...

def setitem(*args, **kwargs) -> None: ...

def sub(*args, **kwargs) -> Any: ...

def truediv(*args, **kwargs) -> Any: ...

def truth(*args, **kwargs) -> bool: ...

def xor(*args, **kwargs) -> Any: ...



# Stubs for redis.utils (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

HIREDIS_AVAILABLE = ... # type: Any

def from_url(url, db=..., **kwargs): ...
def pipeline(redis_obj): ...

class dummy: ...



"""Facilities and constants for generating error messages during type checking.

The type checker itself does not deal with message string literals to
improve code clarity and to simplify localization (in the future)."""

import re
import difflib

from typing import cast, List, Dict, Any, Sequence, Iterable, Tuple

from mypy.errors import Errors
from mypy.types import (
    Type, CallableType, Instance, TypeVarType, TupleType, UnionType, Void, NoneTyp, AnyType,
    Overloaded, FunctionLike, DeletedType
)
from mypy.nodes import (
    TypeInfo, Context, MypyFile, op_methods, FuncDef, reverse_type_aliases,
    ARG_STAR, ARG_STAR2
)


# Constants that represent simple type checker error message, i.e. messages
# that do not have any parameters.

NO_RETURN_VALUE_EXPECTED = 'No return value expected'
INCOMPATIBLE_RETURN_VALUE_TYPE = 'Incompatible return value type'
RETURN_VALUE_EXPECTED = 'Return value expected'
BOOLEAN_VALUE_EXPECTED = 'Boolean value expected'
BOOLEAN_EXPECTED_FOR_IF = 'Boolean value expected for if condition'
BOOLEAN_EXPECTED_FOR_WHILE = 'Boolean value expected for while condition'
BOOLEAN_EXPECTED_FOR_UNTIL = 'Boolean value expected for until condition'
BOOLEAN_EXPECTED_FOR_NOT = 'Boolean value expected for not operand'
INVALID_EXCEPTION = 'Exception must be derived from BaseException'
INVALID_EXCEPTION_TYPE = 'Exception type must be derived from BaseException'
INVALID_RETURN_TYPE_FOR_GENERATOR = \
    'The return type of a generator function should be "Generator" or one of its supertypes'
INVALID_GENERATOR_RETURN_ITEM_TYPE = \
    'The return type of a generator function must be None in its third type parameter in Python 2'
YIELD_VALUE_EXPECTED = 'Yield value expected'
INCOMPATIBLE_TYPES = 'Incompatible types'
INCOMPATIBLE_TYPES_IN_ASSIGNMENT = 'Incompatible types in assignment'
INCOMPATIBLE_REDEFINITION = 'Incompatible redefinition'
INCOMPATIBLE_TYPES_IN_YIELD = 'Incompatible types in yield'
INCOMPATIBLE_TYPES_IN_YIELD_FROM = 'Incompatible types in "yield from"'
INCOMPATIBLE_TYPES_IN_STR_INTERPOLATION = 'Incompatible types in string interpolation'
INIT_MUST_HAVE_NONE_RETURN_TYPE = 'The return type of "__init__" must be None'
GETTER_TYPE_INCOMPATIBLE_WITH_SETTER = \
    'Type of getter incompatible with setter'
TUPLE_INDEX_MUST_BE_AN_INT_LITERAL = 'Tuple index must be an integer literal'
TUPLE_SLICE_MUST_BE_AN_INT_LITERAL = 'Tuple slice must be an integer literal'
TUPLE_INDEX_OUT_OF_RANGE = 'Tuple index out of range'
TYPE_CONSTANT_EXPECTED = 'Type "Constant" or initializer expected'
INCOMPATIBLE_PAIR_ITEM_TYPE = 'Incompatible Pair item type'
INVALID_TYPE_APPLICATION_TARGET_TYPE = 'Invalid type application target type'
INCOMPATIBLE_TUPLE_ITEM_TYPE = 'Incompatible tuple item type'
INCOMPATIBLE_KEY_TYPE = 'Incompatible dictionary key type'
INCOMPATIBLE_VALUE_TYPE = 'Incompatible dictionary value type'
NEED_ANNOTATION_FOR_VAR = 'Need type annotation for variable'
ITERABLE_EXPECTED = 'Iterable expected'
INCOMPATIBLE_TYPES_IN_FOR = 'Incompatible types in for statement'
INCOMPATIBLE_ARRAY_VAR_ARGS = 'Incompatible variable arguments in call'
INVALID_SLICE_INDEX = 'Slice index must be an integer or None'
CANNOT_INFER_LAMBDA_TYPE = 'Cannot infer type of lambda'
CANNOT_INFER_ITEM_TYPE = 'Cannot infer iterable item type'
CANNOT_ACCESS_INIT = 'Cannot access "__init__" directly'
CANNOT_ASSIGN_TO_METHOD = 'Cannot assign to a method'
CANNOT_ASSIGN_TO_TYPE = 'Cannot assign to a type'
INCONSISTENT_ABSTRACT_OVERLOAD = \
    'Overloaded method has both abstract and non-abstract variants'
READ_ONLY_PROPERTY_OVERRIDES_READ_WRITE = \
    'Read-only property cannot override read-write property'
INSTANCE_LAYOUT_CONFLICT = 'Instance layout conflict in multiple inheritance'
FORMAT_REQUIRES_MAPPING = 'Format requires a mapping'
GENERIC_TYPE_NOT_VALID_AS_EXPRESSION = \
    "Generic type is prohibited as a runtime expression (use a type alias or '# type:' comment)"
RETURN_TYPE_CANNOT_BE_CONTRAVARIANT = "Cannot use a contravariant type variable as return type"
FUNCTION_PARAMETER_CANNOT_BE_COVARIANT = "Cannot use a covariant type variable as a parameter"
INCOMPATIBLE_IMPORT_OF = "Incompatible import of"
FUNCTION_TYPE_EXPECTED = "Function is missing a type annotation"
RETURN_TYPE_EXPECTED = "Function is missing a return type annotation"
ARGUMENT_TYPE_EXPECTED = "Function is missing a type annotation for one or more arguments"
KEYWORD_ARGUMENT_REQUIRES_STR_KEY_TYPE = \
    'Keyword argument only valid with "str" key type in call to "dict"'


class MessageBuilder:
    """Helper class for reporting type checker error messages with parameters.

    The methods of this class need to be provided with the context within a
    file; the errors member manages the wider context.

    IDEA: Support a 'verbose mode' that includes full information about types
          in error messages and that may otherwise produce more detailed error
          messages.
    """

    # Report errors using this instance. It knows about the current file and
    # import context.
    errors = None  # type: Errors

    modules = None  # type: Dict[str, MypyFile]

    # Number of times errors have been disabled.
    disable_count = 0

    # Hack to deduplicate error messages from union types
    disable_type_names = 0

    def __init__(self, errors: Errors, modules: Dict[str, MypyFile]) -> None:
        self.errors = errors
        self.modules = modules
        self.disable_count = 0
        self.disable_type_names = 0

    #
    # Helpers
    #

    def copy(self) -> 'MessageBuilder':
        new = MessageBuilder(self.errors.copy(), self.modules)
        new.disable_count = self.disable_count
        new.disable_type_names = self.disable_type_names
        return new

    def add_errors(self, messages: 'MessageBuilder') -> None:
        """Add errors in messages to this builder."""
        if self.disable_count <= 0:
            for info in messages.errors.error_info:
                self.errors.add_error_info(info)

    def disable_errors(self) -> None:
        self.disable_count += 1

    def enable_errors(self) -> None:
        self.disable_count -= 1

    def is_errors(self) -> bool:
        return self.errors.is_errors()

    def report(self, msg: str, context: Context, severity: str, file: str = None) -> None:
        """Report an error or note (unless disabled)."""
        if self.disable_count <= 0:
            self.errors.report(context.get_line(), msg.strip(), severity=severity, file=file)

    def fail(self, msg: str, context: Context, file: str = None) -> None:
        """Report an error message (unless disabled)."""
        self.report(msg, context, 'error', file=file)

    def note(self, msg: str, context: Context, file: str = None) -> None:
        """Report an error message (unless disabled)."""
        self.report(msg, context, 'note', file=file)

    def format(self, typ: Type, verbosity: int = 0) -> str:
        """Convert a type to a relatively short string that is suitable for error messages.

        Mostly behave like format_simple below, but never return an empty string.
        """
        s = self.format_simple(typ, verbosity)
        if s != '':
            # If format_simple returns a non-trivial result, use that.
            return s
        elif isinstance(typ, FunctionLike):
            func = typ
            if func.is_type_obj():
                # The type of a type object type can be derived from the
                # return type (this always works).
                itype = cast(Instance, func.items()[0].ret_type)
                result = self.format(itype)
                if verbosity >= 1:
                    # In some contexts we want to be explicit about the distinction
                    # between type X and the type of type object X.
                    result += ' (type object)'
                return result
            elif isinstance(func, CallableType):
                return_type = strip_quotes(self.format(func.ret_type))
                if func.is_ellipsis_args:
                    return 'Callable[..., {}]'.format(return_type)
                arg_types = [strip_quotes(self.format(t)) for t in func.arg_types]
                return 'Callable[[{}], {}]'.format(", ".join(arg_types), return_type)
            else:
                # Use a simple representation for function types; proper
                # function types may result in long and difficult-to-read
                # error messages.
                return 'overloaded function'
        else:
            # Default case; we simply have to return something meaningful here.
            return 'object'

    def format_simple(self, typ: Type, verbosity: int = 0) -> str:
        """Convert simple types to string that is suitable for error messages.

        Return "" for complex types. Try to keep the length of the result
        relatively short to avoid overly long error messages.

        Examples:
          builtins.int -> 'int'
          Any type -> 'Any'
          None -> None
          callable type -> "" (empty string)
        """
        if isinstance(typ, Instance):
            itype = typ
            # Get the short name of the type.
            if verbosity >= 2:
                base_str = itype.type.fullname()
            else:
                base_str = itype.type.name()
            if itype.args == []:
                # No type arguments. Place the type name in quotes to avoid
                # potential for confusion: otherwise, the type name could be
                # interpreted as a normal word.
                return '"{}"'.format(base_str)
            elif itype.type.fullname() == 'builtins.tuple':
                item_type = strip_quotes(self.format(itype.args[0]))
                return 'Tuple[{}, ...]'.format(item_type)
            elif itype.type.fullname() in reverse_type_aliases:
                alias = reverse_type_aliases[itype.type.fullname()]
                alias = alias.split('.')[-1]
                items = [strip_quotes(self.format(arg)) for arg in itype.args]
                return '{}[{}]'.format(alias, ', '.join(items))
            else:
                # There are type arguments. Convert the arguments to strings
                # (using format() instead of format_simple() to avoid empty
                # strings). If the result is too long, replace arguments
                # with [...].
                a = []  # type: List[str]
                for arg in itype.args:
                    a.append(strip_quotes(self.format(arg)))
                s = ', '.join(a)
                if len((base_str + s)) < 25:
                    return '{}[{}]'.format(base_str, s)
                else:
                    return '{}[...]'.format(base_str)
        elif isinstance(typ, TypeVarType):
            # This is similar to non-generic instance types.
            return '"{}"'.format(typ.name)
        elif isinstance(typ, TupleType):
            # Prefer the name of the fallback class (if not tuple), as it's more informative.
            if typ.fallback.type.fullname() != 'builtins.tuple':
                return self.format_simple(typ.fallback)
            items = []
            for t in typ.items:
                items.append(strip_quotes(self.format(t)))
            s = '"Tuple[{}]"'.format(', '.join(items))
            if len(s) < 40:
                return s
            else:
                return 'tuple(length {})'.format(len(items))
        elif isinstance(typ, UnionType):
            items = []
            for t in typ.items:
                items.append(strip_quotes(self.format(t)))
            s = '"Union[{}]"'.format(', '.join(items))
            if len(s) < 40:
                return s
            else:
                return 'union type ({} items)'.format(len(items))
        elif isinstance(typ, Void):
            return 'None'
        elif isinstance(typ, NoneTyp):
            return 'None'
        elif isinstance(typ, AnyType):
            return '"Any"'
        elif isinstance(typ, DeletedType):
            return '<deleted>'
        elif typ is None:
            raise RuntimeError('Type is None')
        else:
            # No simple representation for this type that would convey very
            # useful information. No need to mention the type explicitly in a
            # message.
            return ''

    def format_distinctly(self, type1: Type, type2: Type) -> Tuple[str, str]:
        """Jointly format a pair of types to distinct strings.

        Increase the verbosity of the type strings until they become distinct.
        """
        verbosity = 0
        for verbosity in range(3):
            str1 = self.format(type1, verbosity=verbosity)
            str2 = self.format(type2, verbosity=verbosity)
            if str1 != str2:
                return (str1, str2)
        return (str1, str2)

    #
    # Specific operations
    #

    # The following operations are for genering specific error messages. They
    # get some information as arguments, and they build an error message based
    # on them.

    def has_no_attr(self, typ: Type, member: str, context: Context) -> Type:
        """Report a missing or non-accessible member.

        The type argument is the base type. If member corresponds to
        an operator, use the corresponding operator name in the
        messages. Return type Any.
        """
        if (isinstance(typ, Instance) and
                typ.type.has_readable_member(member)):
            self.fail('Member "{}" is not assignable'.format(member), context)
        elif isinstance(typ, Void):
            self.check_void(typ, context)
        elif member == '__contains__':
            self.fail('Unsupported right operand type for in ({})'.format(
                self.format(typ)), context)
        elif member in op_methods.values():
            # Access to a binary operator member (e.g. _add). This case does
            # not handle indexing operations.
            for op, method in op_methods.items():
                if method == member:
                    self.unsupported_left_operand(op, typ, context)
                    break
        elif member == '__neg__':
            self.fail('Unsupported operand type for unary - ({})'.format(
                self.format(typ)), context)
        elif member == '__pos__':
            self.fail('Unsupported operand type for unary + ({})'.format(
                self.format(typ)), context)
        elif member == '__invert__':
            self.fail('Unsupported operand type for ~ ({})'.format(
                self.format(typ)), context)
        elif member == '__getitem__':
            # Indexed get.
            self.fail('Value of type {} is not indexable'.format(
                self.format(typ)), context)
        elif member == '__setitem__':
            # Indexed set.
            self.fail('Unsupported target for indexed assignment', context)
        elif member == '__call__':
            self.fail('{} not callable'.format(self.format(typ)), context)
        else:
            # The non-special case: a missing ordinary attribute.
            if not self.disable_type_names:
                failed = False
                if isinstance(typ, Instance) and typ.type.names:
                    alternatives = set(typ.type.names.keys())
                    matches = [m for m in COMMON_MISTAKES.get(member, []) if m in alternatives]
                    matches.extend(best_matches(member, alternatives)[:3])
                    if matches:
                        self.fail('{} has no attribute "{}"; maybe {}?'.format(
                            self.format(typ), member, pretty_or(matches)), context)
                        failed = True
                if not failed:
                    self.fail('{} has no attribute "{}"'.format(self.format(typ),
                                                                member), context)
            else:
                self.fail('Some element of union has no attribute "{}"'.format(
                    member), context)
        return AnyType()

    def unsupported_operand_types(self, op: str, left_type: Any,
                                  right_type: Any, context: Context) -> None:
        """Report unsupported operand types for a binary operation.

        Types can be Type objects or strings.
        """
        if isinstance(left_type, Void) or isinstance(right_type, Void):
            self.check_void(left_type, context)
            self.check_void(right_type, context)
            return
        left_str = ''
        if isinstance(left_type, str):
            left_str = left_type
        else:
            left_str = self.format(left_type)

        right_str = ''
        if isinstance(right_type, str):
            right_str = right_type
        else:
            right_str = self.format(right_type)

        if self.disable_type_names:
            msg = 'Unsupported operand types for {} (likely involving Union)'.format(op)
        else:
            msg = 'Unsupported operand types for {} ({} and {})'.format(
                op, left_str, right_str)
        self.fail(msg, context)

    def unsupported_left_operand(self, op: str, typ: Type,
                                 context: Context) -> None:
        if not self.check_void(typ, context):
            if self.disable_type_names:
                msg = 'Unsupported left operand type for {} (some union)'.format(op)
            else:
                msg = 'Unsupported left operand type for {} ({})'.format(
                    op, self.format(typ))
            self.fail(msg, context)

    def type_expected_as_right_operand_of_is(self, context: Context) -> None:
        self.fail('Type expected as right operand of "is"', context)

    def not_callable(self, typ: Type, context: Context) -> Type:
        self.fail('{} not callable'.format(self.format(typ)), context)
        return AnyType()

    def untyped_function_call(self, callee: CallableType, context: Context) -> Type:
        name = callee.name if callee.name is not None else '(unknown)'
        self.fail('call to untyped function {} in typed context'.format(name), context)
        return AnyType()

    def incompatible_argument(self, n: int, m: int, callee: CallableType, arg_type: Type,
                              arg_kind: int, context: Context) -> None:
        """Report an error about an incompatible argument type.

        The argument type is arg_type, argument number is n and the
        callee type is 'callee'. If the callee represents a method
        that corresponds to an operator, use the corresponding
        operator name in the messages.
        """
        target = ''
        if callee.name:
            name = callee.name
            base = extract_type(name)

            for op, method in op_methods.items():
                for variant in method, '__r' + method[2:]:
                    if name.startswith('"{}" of'.format(variant)):
                        if op == 'in' or variant != method:
                            # Reversed order of base/argument.
                            self.unsupported_operand_types(op, arg_type, base,
                                                           context)
                        else:
                            self.unsupported_operand_types(op, base, arg_type,
                                                           context)
                        return

            if name.startswith('"__getitem__" of'):
                self.invalid_index_type(arg_type, base, context)
                return

            if name.startswith('"__setitem__" of'):
                if n == 1:
                    self.invalid_index_type(arg_type, base, context)
                else:
                    self.fail(INCOMPATIBLE_TYPES_IN_ASSIGNMENT, context)
                return

            target = 'to {} '.format(name)

        msg = ''
        if callee.name == '<list>':
            name = callee.name[1:-1]
            n -= 1
            msg = '{} item {} has incompatible type {}'.format(
                name[0].upper() + name[1:], n, self.format_simple(arg_type))
        elif callee.name == '<list-comprehension>':
            msg = 'List comprehension has incompatible type List[{}]'.format(
                strip_quotes(self.format(arg_type)))
        elif callee.name == '<set-comprehension>':
            msg = 'Set comprehension has incompatible type Set[{}]'.format(
                strip_quotes(self.format(arg_type)))
        elif callee.name == '<dictionary-comprehension>':
            msg = ('{} expression in dictionary comprehension has incompatible type {}; '
                   'expected type {}').format(
                'Key' if n == 1 else 'Value',
                self.format(arg_type),
                self.format(callee.arg_types[n - 1]))
        elif callee.name == '<generator>':
            msg = 'Generator has incompatible item type {}'.format(
                self.format_simple(arg_type))
        else:
            try:
                expected_type = callee.arg_types[m - 1]
            except IndexError:  # Varargs callees
                expected_type = callee.arg_types[-1]
            arg_type_str, expected_type_str = self.format_distinctly(arg_type, expected_type)
            if arg_kind == ARG_STAR:
                arg_type_str = '*' + arg_type_str
            elif arg_kind == ARG_STAR2:
                arg_type_str = '**' + arg_type_str
            msg = 'Argument {} {}has incompatible type {}; expected {}'.format(
                n, target, arg_type_str, expected_type_str)
        self.fail(msg, context)

    def invalid_index_type(self, index_type: Type, base_str: str,
                           context: Context) -> None:
        self.fail('Invalid index type {} for {}'.format(
            self.format(index_type), base_str), context)

    def too_few_arguments(self, callee: CallableType, context: Context,
                          argument_names: List[str]) -> None:
        if (argument_names is not None and not all(k is None for k in argument_names)
                and len(argument_names) >= 1):
            diff = [k for k in callee.arg_names if k not in argument_names]
            if len(diff) == 1:
                msg = 'Missing positional argument'
            else:
                msg = 'Missing positional arguments'
            if callee.name and diff:
                msg += ' "{}" in call to {}'.format('", "'.join(diff), callee.name)
        else:
            msg = 'Too few arguments'
            if callee.name:
                msg += ' for {}'.format(callee.name)
        self.fail(msg, context)

    def too_many_arguments(self, callee: CallableType, context: Context) -> None:
        msg = 'Too many arguments'
        if callee.name:
            msg += ' for {}'.format(callee.name)
        self.fail(msg, context)

    def too_many_positional_arguments(self, callee: CallableType,
                                      context: Context) -> None:
        msg = 'Too many positional arguments'
        if callee.name:
            msg += ' for {}'.format(callee.name)
        self.fail(msg, context)

    def unexpected_keyword_argument(self, callee: CallableType, name: str,
                                    context: Context) -> None:
        msg = 'Unexpected keyword argument "{}"'.format(name)
        if callee.name:
            msg += ' for {}'.format(callee.name)
        self.fail(msg, context)
        if callee.definition:
            fullname = callee.definition.fullname()
            if fullname is not None and '.' in fullname:
                module_name = fullname.rsplit('.', 1)[0]
                path = self.modules[module_name].path
                self.note('{} defined here'.format(callee.name), callee.definition, file=path)

    def duplicate_argument_value(self, callee: CallableType, index: int,
                                 context: Context) -> None:
        self.fail('{} gets multiple values for keyword argument "{}"'.
                  format(capitalize(callable_name(callee)),
                         callee.arg_names[index]), context)

    def does_not_return_value(self, void_type: Type, context: Context) -> None:
        """Report an error about a void type in a non-void context.

        The first argument must be a void type. If the void type has a
        source in it, report it in the error message. This allows
        giving messages such as 'Foo does not return a value'.
        """
        if (cast(Void, void_type)).source is None:
            self.fail('Function does not return a value', context)
        else:
            self.fail('{} does not return a value'.format(
                capitalize((cast(Void, void_type)).source)), context)

    def deleted_as_rvalue(self, typ: DeletedType, context: Context) -> None:
        """Report an error about using an deleted type as an rvalue."""
        if typ.source is None:
            s = ""
        else:
            s = " '{}'".format(typ.source)
        self.fail('Trying to read deleted variable{}'.format(s), context)

    def deleted_as_lvalue(self, typ: DeletedType, context: Context) -> None:
        """Report an error about using an deleted type as an lvalue.

        Currently, this only occurs when trying to assign to an
        exception variable outside the local except: blocks.
        """
        if typ.source is None:
            s = ""
        else:
            s = " '{}'".format(typ.source)
        self.fail('Assignment to variable{} outside except: block'.format(s), context)

    def no_variant_matches_arguments(self, overload: Overloaded, arg_types: List[Type],
                                     context: Context) -> None:
        if overload.name():
            self.fail('No overload variant of {} matches argument types {}'
                      .format(overload.name(), arg_types), context)
        else:
            self.fail('No overload variant matches argument types {}'.format(arg_types), context)

    def function_variants_overlap(self, n1: int, n2: int,
                                  context: Context) -> None:
        self.fail('Function signature variants {} and {} overlap'.format(
            n1 + 1, n2 + 1), context)

    def invalid_cast(self, target_type: Type, source_type: Type,
                     context: Context) -> None:
        if not self.check_void(source_type, context):
            self.fail('Cannot cast from {} to {}'.format(
                self.format(source_type), self.format(target_type)), context)

    def wrong_number_values_to_unpack(self, provided: int, expected: int,
                                      context: Context) -> None:
        if provided < expected:
            if provided == 1:
                self.fail('Need more than 1 value to unpack ({} expected)'.format(expected),
                          context)
            else:
                self.fail('Need more than {} values to unpack ({} expected)'.format(
                    provided, expected), context)
        elif provided > expected:
            self.fail('Too many values to unpack ({} expected, {} provided)'.format(
                expected, provided), context)

    def type_not_iterable(self, type: Type, context: Context) -> None:
        self.fail('\'{}\' object is not iterable'.format(type), context)

    def incompatible_operator_assignment(self, op: str,
                                         context: Context) -> None:
        self.fail('Result type of {} incompatible in assignment'.format(op),
                  context)

    def incompatible_value_count_in_assignment(self, lvalue_count: int,
                                               rvalue_count: int,
                                               context: Context) -> None:
        if rvalue_count < lvalue_count:
            self.fail('Need {} values to assign'.format(lvalue_count), context)
        elif rvalue_count > lvalue_count:
            self.fail('Too many values to assign', context)

    def type_incompatible_with_supertype(self, name: str, supertype: TypeInfo,
                                         context: Context) -> None:
        self.fail('Type of "{}" incompatible with supertype "{}"'.format(
            name, supertype.name), context)

    def signature_incompatible_with_supertype(
            self, name: str, name_in_super: str, supertype: str,
            context: Context) -> None:
        target = self.override_target(name, name_in_super, supertype)
        self.fail('Signature of "{}" incompatible with {}'.format(
            name, target), context)

    def argument_incompatible_with_supertype(
            self, arg_num: int, name: str, name_in_supertype: str,
            supertype: str, context: Context) -> None:
        target = self.override_target(name, name_in_supertype, supertype)
        self.fail('Argument {} of "{}" incompatible with {}'
                  .format(arg_num, name, target), context)

    def return_type_incompatible_with_supertype(
            self, name: str, name_in_supertype: str, supertype: str,
            context: Context) -> None:
        target = self.override_target(name, name_in_supertype, supertype)
        self.fail('Return type of "{}" incompatible with {}'
                  .format(name, target), context)

    def override_target(self, name: str, name_in_super: str,
                        supertype: str) -> str:
        target = 'supertype "{}"'.format(supertype)
        if name_in_super != name:
            target = '"{}" of {}'.format(name_in_super, target)
        return target

    def boolean_return_value_expected(self, method: str,
                                      context: Context) -> None:
        self.fail('Boolean return value expected for method "{}"'.format(
            method), context)

    def incompatible_type_application(self, expected_arg_count: int,
                                      actual_arg_count: int,
                                      context: Context) -> None:
        if expected_arg_count == 0:
            self.fail('Type application targets a non-generic function',
                      context)
        elif actual_arg_count > expected_arg_count:
            self.fail('Type application has too many types ({} expected)'
                      .format(expected_arg_count), context)
        else:
            self.fail('Type application has too few types ({} expected)'
                      .format(expected_arg_count), context)

    def incompatible_array_item_type(self, typ: Type, index: int,
                                     context: Context) -> None:
        self.fail('Array item {} has incompatible type {}'.format(
            index, self.format(typ)), context)

    def could_not_infer_type_arguments(self, callee_type: CallableType, n: int,
                                       context: Context) -> None:
        if callee_type.name and n > 0:
            self.fail('Cannot infer type argument {} of {}'.format(
                n, callee_type.name), context)
        else:
            self.fail('Cannot infer function type argument', context)

    def invalid_var_arg(self, typ: Type, context: Context) -> None:
        self.fail('List or tuple expected as variable arguments', context)

    def invalid_keyword_var_arg(self, typ: Type, context: Context) -> None:
        if isinstance(typ, Instance) and (typ.type.fullname() == 'builtins.dict'):
            self.fail('Keywords must be strings', context)
        else:
            self.fail('Argument after ** must be a dictionary',
                      context)

    def incomplete_type_var_match(self, member: str, context: Context) -> None:
        self.fail('"{}" has incomplete match to supertype type variable'
                  .format(member), context)

    def not_implemented(self, msg: str, context: Context) -> Type:
        self.fail('Feature not implemented yet ({})'.format(msg), context)
        return AnyType()

    def undefined_in_superclass(self, member: str, context: Context) -> None:
        self.fail('"{}" undefined in superclass'.format(member), context)

    def check_void(self, typ: Type, context: Context) -> bool:
        """If type is void, report an error such as '.. does not
        return a value' and return True. Otherwise, return False.
        """
        if isinstance(typ, Void):
            self.does_not_return_value(typ, context)
            return True
        else:
            return False

    def too_few_string_formatting_arguments(self, context: Context) -> None:
        self.fail('Not enough arguments for format string', context)

    def too_many_string_formatting_arguments(self, context: Context) -> None:
        self.fail('Not all arguments converted during string formatting', context)

    def incomplete_conversion_specifier_format(self, context: Context) -> None:
        self.fail('Incomplete format', context)

    def unsupported_placeholder(self, placeholder: str, context: Context) -> None:
        self.fail('Unsupported format character \'%s\'' % placeholder, context)

    def string_interpolation_with_star_and_key(self, context: Context) -> None:
        self.fail('String interpolation contains both stars and mapping keys', context)

    def requires_int_or_char(self, context: Context) -> None:
        self.fail('%c requires int or char', context)

    def key_not_in_mapping(self, key: str, context: Context) -> None:
        self.fail('Key \'%s\' not found in mapping' % key, context)

    def string_interpolation_mixing_key_and_non_keys(self, context: Context) -> None:
        self.fail('String interpolation mixes specifier with and without mapping keys', context)

    def cannot_determine_type(self, name: str, context: Context) -> None:
        self.fail("Cannot determine type of '%s'" % name, context)

    def cannot_determine_type_in_base(self, name: str, base: str, context: Context) -> None:
        self.fail("Cannot determine type of '%s' in base class '%s'" % (name, base), context)

    def invalid_method_type(self, sig: CallableType, context: Context) -> None:
        self.fail('Invalid method type', context)

    def invalid_class_method_type(self, sig: CallableType, context: Context) -> None:
        self.fail('Invalid class method type', context)

    def incompatible_conditional_function_def(self, defn: FuncDef) -> None:
        self.fail('All conditional function variants must have identical '
                  'signatures', defn)

    def cannot_instantiate_abstract_class(self, class_name: str,
                                          abstract_attributes: List[str],
                                          context: Context) -> None:
        attrs = format_string_list("'%s'" % a for a in abstract_attributes[:5])
        self.fail("Cannot instantiate abstract class '%s' with abstract "
                  "attribute%s %s" % (class_name, plural_s(abstract_attributes),
                                   attrs),
                  context)

    def base_class_definitions_incompatible(self, name: str, base1: TypeInfo,
                                            base2: TypeInfo,
                                            context: Context) -> None:
        self.fail('Definition of "{}" in base class "{}" is incompatible '
                  'with definition in base class "{}"'.format(
                      name, base1.name(), base2.name()), context)

    def cant_assign_to_method(self, context: Context) -> None:
        self.fail(CANNOT_ASSIGN_TO_METHOD, context)

    def read_only_property(self, name: str, type: TypeInfo,
                           context: Context) -> None:
        self.fail('Property "{}" defined in "{}" is read-only'.format(
            name, type.name()), context)

    def incompatible_typevar_value(self, callee: CallableType, index: int,
                                   type: Type, context: Context) -> None:
        self.fail('Type argument {} of {} has incompatible value {}'.format(
            index, callable_name(callee), self.format(type)), context)

    def overloaded_signatures_overlap(self, index1: int, index2: int,
                                      context: Context) -> None:
        self.fail('Overloaded function signatures {} and {} overlap with '
                  'incompatible return types'.format(index1, index2), context)

    def invalid_reverse_operator_signature(self, reverse: str, other: str,
                                           context: Context) -> None:
        self.fail('"Any" return type expected since argument to {} '
                  'does not support {}'.format(reverse, other), context)

    def reverse_operator_method_with_any_arg_must_return_any(
            self, method: str, context: Context) -> None:
        self.fail('"Any" return type expected since argument to {} '
                  'has type "Any"'.format(method), context)

    def operator_method_signatures_overlap(
            self, reverse_class: str, reverse_method: str, forward_class: str,
            forward_method: str, context: Context) -> None:
        self.fail('Signatures of "{}" of "{}" and "{}" of "{}" '
                  'are unsafely overlapping'.format(
                      reverse_method, reverse_class,
                      forward_method, forward_class),
                  context)

    def forward_operator_not_callable(
            self, forward_method: str, context: Context) -> None:
        self.fail('Forward operator "{}" is not callable'.format(
            forward_method), context)

    def signatures_incompatible(self, method: str, other_method: str,
                                context: Context) -> None:
        self.fail('Signatures of "{}" and "{}" are incompatible'.format(
            method, other_method), context)

    def yield_from_invalid_operand_type(self, expr: Type, context: Context) -> Type:
        text = self.format(expr) if self.format(expr) != 'object' else expr
        self.fail('"yield from" can\'t be applied to {}'.format(text), context)
        return AnyType()

    def invalid_signature(self, func_type: Type, context: Context) -> None:
        self.fail('Invalid signature "{}"'.format(func_type), context)


def capitalize(s: str) -> str:
    """Capitalize the first character of a string."""
    if s == '':
        return ''
    else:
        return s[0].upper() + s[1:]


def extract_type(name: str) -> str:
    """If the argument is the name of a method (of form C.m), return
    the type portion in quotes (e.g. "y"). Otherwise, return the string
    unmodified.
    """
    name = re.sub('^"[a-zA-Z0-9_]+" of ', '', name)
    return name


def strip_quotes(s: str) -> str:
    """Strip a double quote at the beginning and end of the string, if any."""
    s = re.sub('^"', '', s)
    s = re.sub('"$', '', s)
    return s


def plural_s(s: Sequence[Any]) -> str:
    if len(s) > 1:
        return 's'
    else:
        return ''


def format_string_list(s: Iterable[str]) -> str:
    l = list(s)
    assert len(l) > 0
    if len(l) == 1:
        return l[0]
    else:
        return '%s and %s' % (', '.join(l[:-1]), l[-1])


def callable_name(type: CallableType) -> str:
    if type.name:
        return type.name
    else:
        return 'function'


def temp_message_builder() -> MessageBuilder:
    """Return a message builder usable for throwaway errors (which may not format properly)."""
    return MessageBuilder(Errors(), {})


# For hard-coding suggested missing member alternatives.
COMMON_MISTAKES = {
    'add': ('append', 'extend'),
}  # type: Dict[str, Sequence[str]]


def best_matches(current: str, options: Iterable[str]) -> List[str]:
    ratios = {v: difflib.SequenceMatcher(a=current, b=v).ratio() for v in options}
    return sorted((o for o in options if ratios[o] > 0.75),
                  reverse=True, key=lambda v: (ratios[v], v))


def pretty_or(args: List[str]) -> str:
    quoted = ['"' + a + '"' for a in args]
    if len(quoted) == 1:
        return quoted[0]
    if len(quoted) == 2:
        return "{} or {}".format(quoted[0], quoted[1])
    return ", ".join(quoted[:-1]) + ", or " + quoted[-1]



# Stubs for email.mime.nonmultipart (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from email.mime.base import MIMEBase

class MIMENonMultipart(MIMEBase):
    def attach(self, payload): ...



"""Type parser"""

from typing import List, Tuple, Union, cast, Optional

from mypy.types import (
    Type, UnboundType, TupleType, TypeList, CallableType, StarType,
    EllipsisType
)
from mypy.lex import Token, Name, StrLit, lex
from mypy import nodes


none = Token('')  # Empty token


class TypeParseError(Exception):
    def __init__(self, token: Token, index: int, message: Optional[str] = None) -> None:
        super().__init__()
        self.token = token
        self.index = index
        self.message = message


def parse_type(tok: List[Token], index: int) -> Tuple[Type, int]:
    """Parse a type.

    Return (type, index after type).
    """

    p = TypeParser(tok, index)
    return p.parse_type(), p.index()


def parse_types(tok: List[Token], index: int) -> Tuple[Type, int]:
    """Parse one or more types separated by commas (optional parentheses).

    Return (type, index after type).
    """

    p = TypeParser(tok, index)
    return p.parse_types(), p.index()


class TypeParser:
    def __init__(self, tok: List[Token], ind: int) -> None:
        self.tok = tok
        self.ind = ind

    def index(self) -> int:
        return self.ind

    def parse_type(self) -> Type:
        """Parse a type."""
        t = self.current_token()
        if t.string == '(':
            return self.parse_parens()
        if isinstance(t, Name):
            return self.parse_named_type()
        elif t.string == '[':
            return self.parse_type_list()
        elif t.string == '*':
            return self.parse_star_type()
        elif t.string == '...':
            return self.parse_ellipsis_type()
        elif isinstance(t, StrLit):
            # Type escaped as string literal.
            typestr = t.parsed()
            line = t.line
            self.skip()
            try:
                result = parse_str_as_type(typestr, line)
            except TypeParseError as e:
                raise TypeParseError(e.token, self.ind)
            return result
        else:
            self.parse_error()

    def parse_parens(self) -> Type:
        self.expect('(')
        types = self.parse_types()
        self.expect(')')
        return types

    def parse_types(self) -> Type:
        """ Parse either a single type or a comma separated
        list of types as a tuple type. In the latter case, a
        trailing comma is needed when the list contains only
        a single type (and optional otherwise).

        int   ->   int
        int,  ->   TupleType[int]
        int, int, int  ->  TupleType[int, int, int]
        """
        type = self.parse_type()
        if self.current_token_str() == ',':
            items = [type]
            while self.current_token_str() == ',':
                self.skip()
                if self.current_token_str() == ')':
                    break
                items.append(self.parse_type())
            type = TupleType(items, None, type.line, implicit=True)
        return type

    def parse_type_list(self) -> TypeList:
        """Parse type list [t, ...]."""
        lbracket = self.expect('[')
        commas = []  # type: List[Token]
        items = []  # type: List[Type]
        while self.current_token_str() != ']':
            t = self.parse_type()
            items.append(t)
            if self.current_token_str() != ',':
                break
            commas.append(self.skip())
        self.expect(']')
        return TypeList(items, line=lbracket.line)

    def parse_named_type(self) -> Type:
        line = self.current_token().line
        name = ''
        components = []  # type: List[Token]

        components.append(self.expect_type(Name))
        name += components[-1].string

        while self.current_token_str() == '.':
            components.append(self.skip())
            t = self.expect_type(Name)
            components.append(t)
            name += '.' + t.string

        commas = []  # type: List[Token]
        args = []  # type: List[Type]
        if self.current_token_str() == '[':
            self.skip()
            while True:
                typ = self.parse_type()
                args.append(typ)
                if self.current_token_str() != ',':
                    break
                commas.append(self.skip())

            self.expect(']')

        typ = UnboundType(name, args, line)
        return typ

    def parse_star_type(self) -> Type:
        star = self.expect('*')
        type = self.parse_type()
        return StarType(type, star.line)

    def parse_ellipsis_type(self) -> Type:
        ellipsis = self.expect('...')
        return EllipsisType(ellipsis.line)

    # Helpers

    def skip(self) -> Token:
        self.ind += 1
        return self.tok[self.ind - 1]

    def expect(self, string: str) -> Token:
        if self.tok[self.ind].string == string:
            self.ind += 1
            return self.tok[self.ind - 1]
        else:
            self.parse_error()

    def expect_type(self, typ: type) -> Token:
        if isinstance(self.current_token(), typ):
            self.ind += 1
            return self.tok[self.ind - 1]
        else:
            self.parse_error()

    def current_token(self) -> Token:
        return self.tok[self.ind]

    def current_token_str(self) -> str:
        return self.current_token().string

    def parse_error(self) -> None:
        raise TypeParseError(self.tok[self.ind], self.ind)


def parse_str_as_type(typestr: str, line: int) -> Type:
    """Parse a type represented as a string.

    Raise TypeParseError on parse error.
    """

    typestr = typestr.strip()
    tokens = lex(typestr, line)[0]
    result, i = parse_type(tokens, 0)
    if i < len(tokens) - 2:
        raise TypeParseError(tokens[i], i)
    return result


def parse_str_as_signature(typestr: str, line: int) -> CallableType:
    """Parse a signature represented as a string.

    Raise TypeParseError on parse error.
    """

    typestr = typestr.strip()
    tokens = lex(typestr, line)[0]
    result, i = parse_signature(tokens)
    if i < len(tokens) - 2:
        raise TypeParseError(tokens[i], i)
    return result


def parse_signature(tokens: List[Token]) -> Tuple[CallableType, int]:
    """Parse signature of form (argtype, ...) -> ...

    Return tuple (signature type, token index).
    """
    i = 0
    if tokens[i].string != '(':
        raise TypeParseError(tokens[i], i)
    i += 1
    arg_types = []  # type: List[Type]
    arg_kinds = []  # type: List[int]
    encountered_ellipsis = False
    while tokens[i].string != ')':
        if tokens[i].string == '*':
            arg_kinds.append(nodes.ARG_STAR)
            i += 1
        elif tokens[i].string == '**':
            arg_kinds.append(nodes.ARG_STAR2)
            i += 1
        else:
            arg_kinds.append(nodes.ARG_POS)
        arg, i = parse_type(tokens, i)
        arg_types.append(arg)
        next = tokens[i].string

        # Check for ellipsis. If it exists, assert it's the only arg_type.
        # Disallow '(..., int) -> None' for example.
        if isinstance(arg, EllipsisType):
            encountered_ellipsis = True
        if encountered_ellipsis and len(arg_types) != 1:
            raise TypeParseError(tokens[i], i,
                                 "Ellipses cannot accompany other argument types"
                                 " in function type signature.")

        if next not in ',)':
            raise TypeParseError(tokens[i], i)
        if next == ',':
            i += 1
    i += 1
    if tokens[i].string != '->':
        raise TypeParseError(tokens[i], i)
    i += 1
    ret_type, i = parse_type(tokens, i)
    return CallableType(arg_types,
                        arg_kinds,
                        [None] * len(arg_types),
                        ret_type, None,
                        is_ellipsis_args=encountered_ellipsis), i



# Stubs for base64

# Based on http://docs.python.org/3.2/library/base64.html

from typing import IO

def b64encode(s: str, altchars: str = ...) -> str: ...
def b64decode(s: str, altchars: str = ...,
              validate: bool = ...) -> str: ...
def standard_b64encode(s: str) -> str: ...
def standard_b64decode(s: str) -> str: ...
def urlsafe_b64encode(s: str) -> str: ...
def urlsafe_b64decode(s: str) -> str: ...
def b32encode(s: str) -> str: ...
def b32decode(s: str, casefold: bool = ...,
              map01: str = ...) -> str: ...
def b16encode(s: str) -> str: ...
def b16decode(s: str, casefold: bool = ...) -> str: ...

def decode(input: IO[str], output: IO[str]) -> None: ...
def decodebytes(s: str) -> str: ...
def decodestring(s: str) -> str: ...
def encode(input: IO[str], output: IO[str]) -> None: ...
def encodebytes(s: str) -> str: ...
def encodestring(s: str) -> str: ...



# Stubs for io

# Based on https://docs.python.org/2/library/io.html

# Only a subset of functionality is included.

DEFAULT_BUFFER_SIZE = 0

from typing import List, BinaryIO, TextIO, IO, overload, Iterator, Iterable, Any, Union

def open(file: Union[str, unicode, int],
         mode: unicode = ..., buffering: int = ..., encoding: unicode = ...,
         errors: unicode = ..., newline: unicode = ...,
         closefd: bool = ...) -> IO[Any]: ...

class IOBase:
    # TODO
    ...

class BytesIO(BinaryIO):
    def __init__(self, initial_bytes: str = ...) -> None: ...
    # TODO getbuffer
    # TODO see comments in BinaryIO for missing functionality
    def close(self) -> None: ...
    def closed(self) -> bool: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def read(self, n: int = ...) -> str: ...
    def readable(self) -> bool: ...
    def readline(self, limit: int = ...) -> str: ...
    def readlines(self, hint: int = ...) -> List[str]: ...
    def seek(self, offset: int, whence: int = ...) -> None: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def truncate(self, size: int = ...) -> int: ...
    def writable(self) -> bool: ...
    def write(self, s: str) -> None: ...
    def writelines(self, lines: Iterable[str]) -> None: ...
    def getvalue(self) -> str: ...
    def read1(self) -> str: ...

    def __iter__(self) -> Iterator[str]: ...
    def __enter__(self) -> 'BytesIO': ...
    def __exit__(self, type, value, traceback) -> bool: ...

class StringIO(TextIO):
    def __init__(self, initial_value: unicode = ...,
                 newline: unicode = ...) -> None: ...
    # TODO see comments in BinaryIO for missing functionality
    def close(self) -> None: ...
    def closed(self) -> bool: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def read(self, n: int = ...) -> unicode: ...
    def readable(self) -> bool: ...
    def readline(self, limit: int = ...) -> unicode: ...
    def readlines(self, hint: int = ...) -> List[unicode]: ...
    def seek(self, offset: int, whence: int = ...) -> None: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def truncate(self, size: int = ...) -> int: ...
    def writable(self) -> bool: ...
    def write(self, s: unicode) -> None: ...
    def writelines(self, lines: Iterable[unicode]) -> None: ...
    def getvalue(self) -> unicode: ...

    def __iter__(self) -> Iterator[unicode]: ...
    def __enter__(self) -> 'StringIO': ...
    def __exit__(self, type, value, traceback) -> bool: ...

class TextIOWrapper(TextIO):
    # write_through is undocumented but used by subprocess
    def __init__(self, buffer: IO[str], encoding: unicode = ...,
                 errors: unicode = ..., newline: unicode = ...,
                 line_buffering: bool = ...,
                 write_through: bool = ...) -> None: ...
    # TODO see comments in BinaryIO for missing functionality
    def close(self) -> None: ...
    def closed(self) -> bool: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def read(self, n: int = ...) -> unicode: ...
    def readable(self) -> bool: ...
    def readline(self, limit: int = ...) -> unicode: ...
    def readlines(self, hint: int = ...) -> List[unicode]: ...
    def seek(self, offset: int, whence: int = ...) -> None: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def truncate(self, size: int = ...) -> int: ...
    def writable(self) -> bool: ...
    def write(self, s: unicode) -> None: ...
    def writelines(self, lines: Iterable[unicode]) -> None: ...

    def __iter__(self) -> Iterator[unicode]: ...
    def __enter__(self) -> StringIO: ...
    def __exit__(self, type, value, traceback) -> bool: ...

class BufferedIOBase(IOBase): ...
